from __future__ import annotations

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo import _structs, _util, arrays, compare, sets, text
from mismo.lib.geo._latlon import distance_km
from mismo.lib.geo._regex_parse import parse_street1_re


def _featurize_many(
    t: ibis.Table, *, input_column: str, output_column: str
) -> ibis.Table:
    """Given a table with column addresses:Array<Struct>, add a column address_featured:Array<Struct>."""  # noqa: E501
    with_id = t.mutate(__id=ibis.row_number())
    unnested = with_id.select("__id", _[input_column].unnest().name(input_column))
    lookup = (
        _featurize(
            unnested,
            input_column=input_column,
            output_name="address_featured",
        )
        .group_by("__id")
        .agg(_.address_featured.collect().name(output_column))
    )
    lookup = lookup.cache()
    rejoined = _util.join_lookup(
        with_id, lookup, "__id", defaults={output_column: []}
    ).drop("__id")
    rejoined = rejoined.mutate(
        _[input_column].isnull().ifelse(None, _[output_column]).name(output_column)
    )
    return rejoined


def _norm_field(s: ir.StringValue) -> ir.StringValue:
    return (
        s.strip()
        .upper()
        .re_replace(r"\s+", " ")  # collapse whitespace
        .re_replace(r"[^0-9A-Z\.\- ]", "")
        .nullif("")
    )


def _norm_address(address: ir.StructValue) -> ir.StructValue:
    return ibis.struct(
        {
            "street1": _norm_field(address.street1),
            "street2": _norm_field(address.street2),
            "city": _norm_field(address.city),
            "state": _norm_field(address.state),
            "postal_code": _norm_field(address.postal_code),
        }
    )


def _featurize(
    t: ibis.Table, *, input_column: str | ir.StructValue, output_name: str
) -> ibis.Table:
    """Add a StructColumn named "address_featured" to the table."""
    # we have this Table -> Table API for performance reasons:
    # by doing all these operations in a much of sequential .mutate()s,
    # it compiles to a bunch of chained CTEs in the SQL.
    #
    # If we only did one big .mutate(), some of the deep subexpressions would
    # get copy-pasted many many times (eg Array.unique() results in the argument
    # being copy-pasted 4 times: # https://github.com/ibis-project/ibis/blob/75ecf038d917b968dcc493c69429abe1e8549dd2/ibis/backends/sql/compilers/duckdb.py#L145-L155).
    # If duckdb/the backend were clever with common subexpression elimination,
    # this might not be a problem. But, duckdb appears to evaluate a regex
    # for every time it appears in the SQL. See https://github.com/duckdb/duckdb/discussions/14649.
    # So, if we did one .mutate(), we would end up with like literally 100 regex
    # evaluations in the SQL, which is 100x slower than evaluating the regex once.
    input_column = _util.bind_one(t, input_column)
    t = t.mutate(_parsed=parse_street1_re(input_column.street1), _cleaned=input_column)
    t = t.mutate(
        __address_featured=ibis.struct(
            dict(
                street1=_._cleaned.street1,
                city=_._cleaned.city,
                postal_code=_._cleaned.postal_code,
                state=_._cleaned.state,
                street_name=_norm_field(_._parsed.StreetName),
                # either 123 from "123 Main St" or "PO BOX 123". Only one of these
                # will be non-empty.
                street_number=_norm_field(
                    _._parsed.AddressNumber + "" + _._parsed.USPSBoxID
                ),
                is_pobox=_._parsed.USPSBoxID != "",
            )
        )
    )
    t = t.mutate(
        __address_featured=_structs.mutate(
            t.__address_featured,
            street_number_sorted=t.__address_featured.street_number.split("")
            .sort()
            .join(""),
        )
    )
    t = t.mutate(
        _structs.mutate(
            t.__address_featured,
            street_trigrams=text.ngrams(t._cleaned.street1, 3).unique(),
        ).name(output_name)
    ).drop(
        # "_parsed",
        "__address_featured",
    )
    return t


class AddressesMatchLevel(compare.MatchLevel):
    """How closely two addresses match."""

    STREET1_AND_CITY_OR_POSTAL = 0
    """The street1, city, and state match."""
    POSSIBLE_TYPO = 1
    """If you consider typos, the addresses match.
    
    eg the levenstein distance is below a certain threshold.
    """
    SAME_REGION = 2
    """The postal code, or city and state, match."""
    WITHIN_100KM = 3
    """The addresses are within 100 km of each other."""
    SAME_STATE = 4
    """The states match."""
    ELSE = 6
    """None of the above."""


def _is_possible_typo(left: ir.StructValue, right: ir.StructValue) -> ir.BooleanValue:
    cases = [
        ibis.and_(
            left.street1 == right.street1,
            left.street1.length() >= 5,
        ),
        ibis.and_(
            text.damerau_levenshtein_ratio(left.street_name, right.street_name) > 0.9,
            text.damerau_levenshtein_ratio(left.city, right.city) > 0.9,
        ),
        ibis.and_(
            text.damerau_levenshtein_ratio(left.street_name, right.street_name) > 0.9,
            # >=.8 so that a transposition in a 5digit zipcode match,
            # eg "12345" and "12354"
            text.damerau_levenshtein_ratio(left.postal_code, right.postal_code) >= 0.8,
        ),
        ibis.and_(
            left.street_number == right.street_number,
            text.damerau_levenshtein_ratio(left.street_name, right.street_name) > 0.4,
            text.damerau_levenshtein_ratio(left.city, right.city) > 0.9,
        ),
    ]
    return ibis.or_(*cases)


def match_level(left: ir.StructValue, right: ir.StructValue) -> ir.IntegerValue:
    """Compare two address structs, and return the match level."""
    if "latitude" in left.type().names:
        within_100km_levels = [
            (
                distance_km(
                    lat1=left.latitude,
                    lon1=left.longitude,
                    lat2=right.latitude,
                    lon2=right.longitude,
                )
                <= 100,
                AddressesMatchLevel.WITHIN_100KM.as_integer(),
            ),
        ]
    else:
        within_100km_levels = []

    return _util.cases(
        (
            ibis.and_(
                left.street1 == right.street1,
                ibis.or_(
                    left.city == right.city, left.postal_code == right.postal_code
                ),
            ),
            AddressesMatchLevel.STREET1_AND_CITY_OR_POSTAL.as_integer(),
        ),
        (
            _is_possible_typo(left, right),
            AddressesMatchLevel.POSSIBLE_TYPO.as_integer(),
        ),
        (
            ibis.or_(
                left.postal_code == right.postal_code,
                ibis.and_(left.city == right.city, left.state == right.state),
            ),
            AddressesMatchLevel.SAME_REGION.as_integer(),
        ),
        *within_100km_levels,
        (left.state == right.state, AddressesMatchLevel.SAME_STATE.as_integer()),
        else_=AddressesMatchLevel.ELSE.as_integer(),
    )


class AddressesDimension:
    """Preps, blocks, and compares based on array<address> columns.

    An address is a Struct of the type
    `struct<
        street1: string,
        street2: string,  # eg "Apt 3"
        city: string,
        state: string,
        postal_code: string,  # zipcode in the US
        country: string,
    >`.
    This operates on columns of type `array<address>`. In other words,
    it is useful for comparing eg people, who might have multiple addresses,
    and they are the same person if any of their addresses match.
    """

    def __init__(
        self,
        column: str,
        *,
        column_cleaned: str = "{column}_cleaned",
        column_featured: str = "{column}_featured",
        column_compared: str = "{column}_compared",
    ):
        self.column = column
        self.column_cleaned = column_cleaned.format(column=column)
        self.column_featured = column_featured.format(column=column)
        self.column_compared = column_compared.format(column=column)

    def prepare_for_fast_linking(self, t: ir.Table) -> ir.Table:
        """Prepares the table for fast linking, adding a normalized column."""
        t = t.mutate(t[self.column].map(_norm_address).name(self.column_cleaned))
        return t

    def prepare_for_blocking(self, t: ir.Table) -> ir.Table:
        """Prepares the table for blocking, adding normalized and tokenized columns."""
        t = _featurize_many(
            t, input_column=self.column_cleaned, output_column=self.column_featured
        )
        t = t.cache()
        t = t.mutate(
            _addresses_tokens=_.addresses_featured.map(lambda a: a.street_trigrams)
            .flatten()
            .unique()
        )
        t = t.cache()
        t = arrays.array_filter_isin_other(
            t,
            t._addresses_tokens,
            sets.rare_terms(t._addresses_tokens, max_records_n=500),
            result_format="_addresses_keywords",
        )
        t = t.cache()
        t = t.mutate(
            ibis.struct(
                {
                    "addresses": t.addresses_featured,
                    "addresses_keywords": _._addresses_keywords,
                }
            ).name(self.column_featured)
        ).drop("_addresses_tokens", "_addresses_keywords")
        t = t.cache()
        return t

    def compare(self, t: ir.Table) -> ir.Table:
        left = t[self.column_featured + "_l"].addresses
        right = t[self.column_featured + "_r"].addresses
        combos = arrays.array_combinations(left, right)
        levels = combos.map(lambda pair: match_level(pair.l, pair.r))
        best = arrays.array_min(levels)
        return t.mutate(best.name(self.column_compared))



from __future__ import annotations

import ibis
import pandas as pd
import pytest

from mismo import vector


def _to_vector(x):
    if isinstance(x, dict):
        return ibis.literal(x, type="map<string, int64>")
    if isinstance(x, list):
        return ibis.literal(x, type="array<int64>")
    return x


@pytest.mark.parametrize(
    "a,b,expected",
    [
        pytest.param([1, 2], [4, 5], 14.0, id="array_int"),
        pytest.param([1.0, 2.0], [4, 5], 14.0, id="array_float"),
        pytest.param({"a": 1, "b": 2}, {"b": 3, "c": 4}, 6.0, id="map"),
        pytest.param([], [], 0, id="empty_arrays"),
        pytest.param({}, {}, 0, id="empty_maps"),
        pytest.param({}, {"a": 5}, 0, id="one_empty_map"),
        pytest.param(
            [1],
            [4, 5],
            9999,
            id="array_length_mismatch",
            marks=pytest.mark.xfail(reason="fails at execution time"),
        ),
        pytest.param(
            ibis.literal(None, "map<string, int64>"),
            {"a": 5},
            None,
            id="null_nonull_map",
        ),
        pytest.param(
            {"a": 5},
            ibis.literal(None, "map<string, int64>"),
            None,
            id="nonnull_null_map",
        ),
        pytest.param(
            ibis.literal(None, "array<int64>"),
            [5, 3],
            None,
            id="one_null_array",
        ),
        pytest.param(
            ibis.literal(None, "array<int64>"),
            ibis.literal(None, "array<int64>"),
            None,
            id="both_null_array",
        ),
        pytest.param(
            ibis.literal(None, "map<string, int64>"),
            ibis.literal(None, "map<string, int64>"),
            None,
            id="both_null_map",
        ),
    ],
)
def test_dot(a, b, expected):
    result = vector.dot(_to_vector(a), _to_vector(b)).execute()
    if pd.isna(result):
        result = None
    assert result == expected


@pytest.mark.parametrize(
    "a,b,expected",
    [
        pytest.param([1, 1], [-2, -2], -1.0, id="opposite_directions"),
        pytest.param([1, 2], [2, 4], 1.0, id="same_directions"),
        pytest.param([1], [-99], -1.0, id="1D_arrays"),
        pytest.param(
            [],
            [],
            0.0,
            id="empty_arrays",
            marks=pytest.mark.xfail(
                reason="https://github.com/duckdb/duckdb/issues/12960"
            ),
        ),
        pytest.param(
            {},
            {},
            0.0,
            id="empty_maps",
            marks=pytest.mark.xfail(
                reason="https://github.com/duckdb/duckdb/issues/12960"
            ),
        ),
        pytest.param(
            {},
            {"a": 5},
            0.0,
            id="one_empty_map",
            marks=pytest.mark.xfail(
                reason="https://github.com/duckdb/duckdb/issues/12960"
            ),
        ),
        pytest.param({"a": 1, "b": 1}, {"a": -2, "b": -2}, -1.0, id="opposite_maps"),
        pytest.param(
            ibis.literal(None, "array<int64>"),
            [5, 3],
            None,
            id="one_null_array",
        ),
        pytest.param(
            [5, 3],
            ibis.literal(None, "array<int64>"),
            None,
            id="nonnull_null_array",
        ),
        pytest.param(
            ibis.literal(None, "array<int64>"),
            ibis.literal(None, "array<int64>"),
            None,
            id="both_null_array",
        ),
    ],
)
def test_cosine_similarity(a, b, expected):
    def to_ibis(x):
        if isinstance(x, dict):
            return ibis.literal(x, type="map<string, int64>")
        if isinstance(x, list):
            return ibis.literal(x, type="array<int64>")
        return x

    a = to_ibis(a)
    b = to_ibis(b)
    e = vector.cosine_similarity(a, b)
    result = e.execute()
    if pd.isna(result):
        result = None
    assert result == pytest.approx(expected)


@pytest.mark.parametrize(
    "a,metric,expected",
    [
        pytest.param([-3, 4], "l2", 5.0, id="array_l2"),
        pytest.param([3, 4], "l1", 7, id="array_l1"),
        pytest.param({"a": -3, "b": 4}, "l2", 5.0, id="map_l2"),
        pytest.param({"a": -3, "b": 4}, "l1", 7, id="map_l1"),
        pytest.param([], "l2", None, id="array_empty"),
        pytest.param({}, "l2", None, id="map_empty"),
        pytest.param(ibis.literal(None, "array<int64>"), "l2", None, id="null_array"),
        pytest.param(
            ibis.literal(None, "map<string, int64>"), "l1", None, id="null_map_l1"
        ),
        pytest.param(
            ibis.literal(None, "map<string, int64>"), "l2", None, id="null_map_l2"
        ),
    ],
)
def test_norm(a, metric, expected):
    e = vector.norm(_to_vector(a), metric=metric)
    result = e.execute()
    if pd.isna(result):
        result = None
    assert result == expected


@pytest.mark.parametrize(
    "a,metric,expected",
    [
        pytest.param(
            [-1, 2], "l2", [-0.4472135954999579, 0.8944271909999159], id="array_l2"
        ),
        pytest.param(
            [-1, 2], "l1", [-0.3333333333333333, 0.6666666666666666], id="array_l1"
        ),
        pytest.param(
            {"a": 1, "b": 2},
            "l2",
            {"a": 0.4472135954999579, "b": 0.8944271909999159},
            id="map_l2",
        ),
        pytest.param(
            {"a": 1, "b": 2},
            "l1",
            {"a": 0.3333333333333333, "b": 0.6666666666666666},
            id="map_l1",
        ),
        pytest.param([], "l2", [], id="array_empty"),
        pytest.param({}, "l2", {}, id="map_empty"),
        pytest.param(ibis.literal(None, "array<int64>"), "l2", None, id="null_array"),
        pytest.param(
            ibis.literal(None, "map<string, int64>"), "l1", None, id="null_map_l1"
        ),
        pytest.param(
            ibis.literal(None, "map<string, int64>"), "l2", None, id="null_map_l2"
        ),
    ],
)
def test_normalize(a, metric, expected):
    result = vector.normalize(_to_vector(a), metric=metric)
    assert result.execute() == expected


@pytest.mark.parametrize(
    "a,b,expected",
    [
        pytest.param([1, 2], [4, 5], [4, 10], id="array_int"),
        pytest.param([1.0, 2.0], [4, 5], [4.0, 10.0], id="array_float"),
        pytest.param([], [], [], id="array_empty"),
        pytest.param(
            [1],
            [4, 5],
            9999,
            id="array_length_mismatch",
            marks=pytest.mark.xfail(reason="fails at execution time"),
        ),
        pytest.param(
            ibis.literal(None, "array<int64>"),
            [5, 3],
            None,
            id="array_nonnull_null",
        ),
        pytest.param(
            [5, 3],
            ibis.literal(None, "array<int64>"),
            None,
            id="array_null_nonnull",
        ),
        pytest.param(
            ibis.literal(None, "array<int64>"),
            ibis.literal(None, "array<int64>"),
            None,
            id="array_both_null",
        ),
        pytest.param({"a": 1, "b": 2}, {"b": 3, "c": 4}, {"b": 6}, id="map_int"),
        pytest.param(
            {"a": 1.0, "b": 2.0}, {"b": 3.0, "c": 4.0}, {"b": 6.0}, id="map_float"
        ),
        pytest.param({}, {}, {}, id="map_empty"),
        pytest.param({}, {"a": 5}, {}, id="map_empty_nonempty"),
        pytest.param(
            ibis.literal(None, "map<string, int64>"),
            {"a": 5},
            None,
            id="map_null_nonull",
        ),
        pytest.param(
            {"a": 5},
            ibis.literal(None, "map<string, int64>"),
            None,
            id="map_nonnull_null",
        ),
        pytest.param(
            ibis.literal(None, "map<string, int64>"),
            ibis.literal(None, "map<string, int64>"),
            None,
            id="map_both_null",
        ),
    ],
)
def test_mul(a, b, expected):
    result = vector.mul(_to_vector(a), _to_vector(b)).execute()
    if expected is None:
        assert pd.isna(result)
    else:
        assert result == expected



from __future__ import annotations

import ibis

from mismo.tests.util import assert_equal
from mismo.types._wrapper import StructWrapper, TableWrapper


def test_table_wrapper_basic(table_factory):
    table = table_factory({"a": [1, 2], "b": [3, 4], "_x": [5, 6]})
    wrapped = TableWrapper(table)
    assert isinstance(wrapped, TableWrapper)
    assert isinstance(wrapped, ibis.Table)
    assert wrapped.schema() == table.schema()
    assert wrapped.columns == table.columns
    assert_equal(wrapped.a, table.a)
    assert_equal(wrapped.b, table.b)
    assert_equal(wrapped["_x"], table["_x"])


def test_struct_wrapper_basic(table_factory):
    table = table_factory({"a": [1, 2], "_x": [5, 6]})
    struct_col = ibis.struct({"a": table.a, "b": 5, "_x": table["_x"]})
    wrapped = StructWrapper(struct_col)
    assert isinstance(wrapped, StructWrapper)
    assert isinstance(wrapped, ibis.Value)
    assert wrapped.type() == struct_col.type()
    assert wrapped.fields == struct_col.fields
    assert_equal(wrapped.a, struct_col.a)
    assert_equal(wrapped.b, struct_col.b)
    assert_equal(wrapped["_x"], struct_col["_x"])
    assert wrapped.to_list() == struct_col.to_list()


def test_table_wrapper_custom_method(table_factory):
    class CustomTableWrapper(TableWrapper):
        def custom_method(self) -> str:
            return "custom_method_called"

        # This should override the underlying table's select method
        def select(self, *args, **kwargs) -> str:
            return "select_called"

    table = table_factory({"a": [1, 2], "b": [3, 4]})
    wrapped = CustomTableWrapper(table)
    assert isinstance(wrapped, CustomTableWrapper)
    assert isinstance(wrapped, TableWrapper)
    assert isinstance(wrapped, ibis.Table)
    assert wrapped.custom_method() == "custom_method_called"
    assert wrapped.select("a") == "select_called"


def test_struct_wrapper_custom_method(table_factory):
    class CustomStructWrapper(StructWrapper):
        def custom_method(self) -> str:
            return "custom_method_called"

        # This should override the underlying struct's isnull method
        def isnull(self) -> str:
            return "isnull_called"

    table = table_factory({"a": [1, 2], "b": [3, 4]})
    struct_col = ibis.struct({"a": table.a, "b": 5})
    wrapped = CustomStructWrapper(struct_col)
    assert isinstance(wrapped, CustomStructWrapper)
    assert isinstance(wrapped, StructWrapper)
    assert isinstance(wrapped, ibis.Value)
    assert wrapped.custom_method() == "custom_method_called"
    assert wrapped.isnull() == "isnull_called"



from __future__ import annotations

from typing import Literal

import ibis

from mismo import joins
from mismo.linkage._linkage import Linkage
from mismo.linker._common import Linker, infer_task
from mismo.types import LinksTable


class JoinLinker(Linker):
    """
    A [Linker][mismo.Linker] based on a join condition.
    """  # noqa: E501

    condition: joins.HasJoinCondition
    task: Literal["dedupe", "link"] | None
    on_slow: Literal["error", "warn", "ignore"]

    def __init__(
        self,
        condition: joins.IntoJoinCondition,
        *,
        task: Literal["dedupe", "link"] | None = None,
        on_slow: Literal["error", "warn", "ignore"] = "error",
    ) -> None:
        """Create from a join condition.

        Parameters
        ----------
        condition
            The join condition.
        task
            The task to perform. If `None`, the task will be inferred as
            "dedupe" if the two tables passed to __call__ are the same,
            otherwise it will be inferred as "link".
        on_slow
            What to do if the join is slow. See [mismo.joins.check_join_algorithm][].
        """
        self.condition = joins.join_condition(condition)
        self.task = task
        self.on_slow = on_slow

    def __call__(self, left: ibis.Table, right: ibis.Table) -> Linkage:
        """Create a linkage from the join condition."""
        task = infer_task(task=self.task, left=left, right=right)
        if left is right:
            right = right.view()
        # Run this to check early for slow joins
        self.__join_condition__(left, right)
        condition = self.__join_condition__(left, right)
        if task == "dedupe":
            condition = condition & (left.record_id < right.record_id)
        links = LinksTable.from_join_condition(left, right, condition)
        return Linkage(left=left, right=right, links=links)

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> bool | ibis.ir.BooleanValue:
        """Create a join condition."""
        pred = self.condition.__join_condition__(left, right)
        joins.check_join_algorithm(left, right, pred, on_slow=self.on_slow)
        return pred

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<condition={self.condition!r}, task={self.task}, on_slow={self.on_slow}>"  # noqa: E501



from __future__ import annotations

import re

import ibis.backends.sql.compilers as sc
from ibis.common.annotations import attribute
from ibis.expr import datatypes as dt
from ibis.expr import operations as ops
from ibis.expr import rules as rlz
from ibis.expr import types as ir
import sqlglot.expressions as sge


def re_extract_struct(
    s: ir.StringValue, pattern: str, *, case_insensitive: bool = False
) -> ir.StructValue:
    r"""Use a regex with named groups to extract parts of a string into a struct.

    Examples
    --------
    >>> import ibis
    >>> s = ibis.literal("2024-03-01")
    >>> pattern = r"(?P<year>\d+)-(?P<month>\d+)-(?P<day>\d+)"
    >>> re_extract_struct(s, pattern).execute()
    {'year': '2024', 'month': '03', 'day': '01'}
    """
    return RegexExtractStruct(s, pattern, case_insensitive=case_insensitive).to_expr()


class RegexExtractStruct(ops.Value):
    """Use a regex to extract parts of a string into a struct"""

    arg: ops.Value[dt.String]
    pattern: str
    case_insensitive: bool = False

    shape = rlz.shape_like("arg")

    @attribute
    def group_names(self) -> list[str]:
        group_name_pattern = r"\(\?P<([^>]+)>"
        return re.findall(group_name_pattern, self.pattern)

    @attribute
    def dtype(self):
        return dt.Struct.from_tuples([(name, dt.string) for name in self.group_names])


def visit_RegexExtractStruct(self, op: RegexExtractStruct, *, arg, pattern, **kwargs):
    flags = "i" if op.case_insensitive else "c"
    try:
        # Ibis changed their internal API, workaround that here.
        return self.f.regexp_extract(
            arg,
            pattern,
            sge.convert(op.group_names),
            flags,
            dialect=self.dialect,
        )
    except TypeError as e:
        if "got multiple values for keyword argument 'dialect'" not in str(e):
            raise
    return self.f.regexp_extract(
        arg,
        pattern,
        sge.convert(op.group_names),
        flags,
    )


# monkeypatch that sucker in
sc.DuckDBCompiler.visit_RegexExtractStruct = visit_RegexExtractStruct



from __future__ import annotations

from mismo.joins._analyze import JOIN_ALGORITHMS as JOIN_ALGORITHMS
from mismo.joins._analyze import SLOW_JOIN_ALGORITHMS as SLOW_JOIN_ALGORITHMS
from mismo.joins._analyze import JoinAlgorithm as JoinAlgorithm
from mismo.joins._analyze import SlowJoinAlgorithm as SlowJoinAlgorithm
from mismo.joins._analyze import check_join_algorithm as check_join_algorithm
from mismo.joins._analyze import get_join_algorithm as get_join_algorithm
from mismo.joins._conditions import AndJoinCondition as AndJoinCondition
from mismo.joins._conditions import BinaryFuncJoinCondition as BinaryFuncJoinCondition
from mismo.joins._conditions import BooleanJoinCondition as BooleanJoinCondition
from mismo.joins._conditions import HasJoinCondition as HasJoinCondition
from mismo.joins._conditions import IntoJoinCondition as IntoJoinCondition
from mismo.joins._conditions import KeyJoinCondition as KeyJoinCondition
from mismo.joins._conditions import join_condition as join_condition
from mismo.joins._conditions import left as left
from mismo.joins._conditions import right as right
from mismo.joins._core import join as join
from mismo.joins._core import remove_condition_overlap as remove_condition_overlap



from __future__ import annotations

import pytest


@pytest.fixture
def name_table(table_factory):
    records = [
        {
            "record_id": "<null>",
            "name": None,
        },
        {
            "record_id": "alice_anderson",
            "name": {
                "given": "Alice",
                "surname": "Anderson",
            },
        },
        {
            "record_id": "a_anderson_phd",
            "name": {
                "given": "A",
                "surname": "Anderson",
                "suffix": "PhD",
            },
        },
        {
            "record_id": "bob_baker",
            "name": {
                "given": "Bob",
                "surname": "Baker",
            },
        },
        {
            "record_id": "robert_b_baker_jr",
            "name": {
                "given": "Robert",
                "middle": "b",
                "surname": "Baker",
                "suffix": "Jr.",
            },
        },
        {
            "record_id": "mr_charles_carter",
            "name": {
                "prefix": "Mr",
                "given": "Charles",
                "surname": "Carter",
                "nickname": "Charlie",
            },
        },
        {
            "record_id": "charles_carter",
            "name": {
                "given": "  CHARLES",
                "surname": " CARTER.",
            },
        },
    ]
    base = {
        "prefix": None,
        "given": None,
        "middle": None,
        "surname": None,
        "suffix": None,
        "nickname": None,
    }
    records = [
        {
            **r,
            "name": {**base, **(r["name"])} if r["name"] is not None else None,
        }
        for r in records
    ]
    return table_factory(records)



from __future__ import annotations

import colorsys
from itertools import product
from typing import TYPE_CHECKING, Iterable

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo.compare._match_level import LevelComparer
from mismo.fs._plot import log_odds_color_scale
from mismo.fs._util import odds_to_log_odds
from mismo.fs._weights import Weights

if TYPE_CHECKING:
    import altair as alt
    import ipywidgets
    import pandas as pd


def compared_dashboard(
    compared: ibis.Table,
    comparers: Iterable[LevelComparer],
    weights: Weights | None = None,
    *,
    width: int = 500,
) -> ipywidgets.VBox:
    """A dashboard for debugging compared record pairs.

    Used to see which match levels are common, which are rare, and which
    Comparers are related to each other. For example, exact matches should
    appear together across all Comparers, this probably represents true
    matches.

    Parameters
    ----------
    compared :
        The result of running the blocked table through the supplied `comparers`.
    comparers :
        The LevelCompareres that were used to compare `compared`.
    weights :
        The Weights used to score the comparers.
        If provided, the chart will be colored by the odds found from the Weights.
    width :
        The width of the chart.

    Returns
    -------
    ipywidgets.VBox
        The dashboard.
    """
    import altair as alt
    import ipywidgets

    cols = [comp.name for comp in comparers]
    compared = compared.mutate(
        vector_id=ibis.literal(":").join(
            [c.levels(compared[c.name]).as_string() for c in comparers]
        ),
    )
    chart = _compared_chart(compared, comparers, weights, width=width)
    chart_widget = alt.JupyterChart(chart)
    limiter_widget = ipywidgets.IntSlider(
        min=1, max=100, value=5, step=1, description="Show Pairs:"
    )
    vector_title_widget = ipywidgets.HTML()
    table_widget = ipywidgets.HTML()

    def set_table(data: ir.Table):
        df = data.limit(limiter_widget.value).to_pandas()
        table_widget.value = df.to_html(index=False, render_links=True)

    set_table(compared.head(0))

    def on_select(change):
        vector_id = change.new.value[0]["vector_id"]
        filtered = compared[_.vector_id == vector_id]
        vector_values = filtered[cols].limit(1).to_pandas().iloc[0].to_dict()
        vector_title_widget.value = "&nbsp;".join(
            f"<b>{k}</b>: {v}" for k, v in vector_values.items()
        )
        set_table(filtered.drop("vector_id", *cols))

    chart_widget.selections.observe(on_select, ["vector_click"])

    return ipywidgets.VBox(
        [
            chart_widget,
            limiter_widget,
            vector_title_widget,
            table_widget,
        ]
    )


def _compared_chart(
    compared: ir.Table,
    comparers: Iterable[LevelComparer],
    weights: Weights | None = None,
    *,
    width: int = 500,
) -> alt.Chart:
    import altair as alt

    cols = [comp.name for comp in comparers]

    vector_counts = compared.group_by(cols + ["vector_id"]).agg(n_pairs=_.count())
    vector_counts = vector_counts.mutate(
        pct_pairs=_.n_pairs / _.n_pairs.sum(),
    )
    if weights is not None:
        vector_counts = weights.score_compared(vector_counts)
        vector_counts = vector_counts.mutate(odds=_.odds.clip(upper=10**10))
        vector_counts = vector_counts.mutate(log_odds=odds_to_log_odds(_.odds))
        hist_color = alt.Color(
            "log_odds",
            title="Odds",
            scale=log_odds_color_scale(),
            legend=alt.Legend(labelExpr=10**alt.datum.value),
        )
        hist_extra_tooltips = [alt.Tooltip("odds", title="Odds", format=",")]
    else:
        hist_color = None
        hist_extra_tooltips = []
    vector_counts = vector_counts.to_pandas()

    scrubber_filter = alt.selection_interval(encodings=["x"])
    vector_fader_mouseover = alt.selection_point(
        name="vector_mouseover", fields=["vector_id"], on="mouseover"
    )
    vector_fader_click = alt.selection_point(
        name="vector_click", fields=["vector_id"], on="click"
    )
    opacity_vector = alt.condition(vector_fader_mouseover, alt.value(1), alt.value(0.8))
    level_color = alt.Color(
        "level_uid",
        title="Comparer:Level",
        scale=_make_level_color_scale(comparers),
        legend=None,
    )
    x = alt.X(
        "vector_id:N",
        axis=None,
        sort=alt.EncodingSortField(
            "n_pairs" if weights is None else "odds", order="descending"
        ),
    )
    scrubber_chart = (
        alt.Chart(
            vector_counts,
            height=50,
            width=width,
            title=alt.Title(
                text="<drag to select>",
                dy=20,
                anchor="middle",
                fontSize=12,
                color="gray",
            ),
        )
        .mark_rect()
        .encode(
            x=x,
            y=alt.Y(
                "n_pairs:Q",
                title="Number of Pairs",
                scale=alt.Scale(type="log"),
                axis=None,
            ),
            opacity=opacity_vector,
            **{"color": hist_color.legend(None)} if hist_color is not None else {},
        )
        .add_params(scrubber_filter)
        .add_params(vector_fader_mouseover)
    )
    hist = (
        alt.Chart(vector_counts, width=width)
        .mark_rect(stroke=None)
        .encode(
            x=x,
            y=alt.Y(
                "n_pairs:Q",
                title="Number of Pairs",
                scale=alt.Scale(type="log"),
            ),
            opacity=opacity_vector,
            tooltip=[
                alt.Tooltip("n_pairs", title="Number of Pairs", format=","),
                alt.Tooltip("pct_pairs", title="Percent of Pairs", format="%"),
                *hist_extra_tooltips,
                *cols,
            ],
            **{"color": hist_color} if hist_color is not None else {},
        )
        .transform_filter(scrubber_filter)
        .add_params(vector_fader_mouseover)
        .add_params(vector_fader_click)
    )
    vector_chart = (
        alt.Chart(_vector_grid_data(comparers, vector_counts), height=80, width=width)
        .mark_rect()
        .encode(
            x=x,
            y=alt.Y(
                "comparer",
                title="Comparer",
                sort=cols,
            ),
            color=level_color,
            opacity=opacity_vector,
            tooltip=["level"],
        )
        .transform_filter(scrubber_filter)
        .add_params(vector_fader_mouseover)
        .add_params(vector_fader_click)
    )
    together = alt.vconcat(scrubber_chart, hist, vector_chart, spacing=0)
    together = together.properties(
        title=alt.Title(
            text="Distribution of Match Levels",
            subtitle=f"Total Pairs: {vector_counts.n_pairs.sum():,}",
            anchor="middle",
            fontSize=14,
        )
    )
    return together


def _frange(start, stop, n):
    return [start + i * (stop - start) / n for i in range(n)]


def _vector_grid_data(
    comparers: Iterable[LevelComparer], vector_data: pd.DataFrame
) -> pd.DataFrame:
    import pandas as pd

    records = []
    for levels in product(*(c.levels for c in comparers)):
        vector_id = ":".join(levels)
        for comp, level in zip(comparers, levels):
            level_info = {c.name: None for c in comparers}
            level_info[comp.name] = level
            records.append(
                {
                    "vector_id": vector_id,
                    "level_uid": _level_uid(comp, level),
                    "comparer": comp.name,
                    "level": level,
                    **level_info,
                }
            )
    result = pd.DataFrame(records)
    # only include vectors that are in the vector_data
    # and add in per-vector info such as n_pairs
    result = vector_data.merge(result, on="vector_id", how="left")
    return result


def _make_level_color_scale(comparers: Iterable[LevelComparer]) -> alt.Scale:
    import altair as alt

    domain = []
    range = []
    hues = _frange(0, 1, len(comparers))
    for comp, hue in zip(comparers, hues):
        levels = comp.levels
        shades = _frange(0.9, 0.2, len(levels))
        for level_name, shade in zip(levels, shades):
            r, g, b = colorsys.hsv_to_rgb(hue, 1, shade)
            r = int(r * 255)
            g = int(g * 255)
            b = int(b * 255)
            hex_color = f"#{r:02x}{g:02x}{b:02x}"
            domain.append(_level_uid(comp, level_name))
            range.append(hex_color)
    return alt.Scale(domain=domain, range=range)


def _level_uid(comparer: LevelComparer, level: str) -> str:
    return comparer.name + ":" + level


# TODO: make this work as a filter for the above histogram
def _make_legend_plot(longer: ir.Table, color_map):
    import altair as alt

    levels = longer.group_by(["comparer", "level"]).agg(
        id=_.id.first(),
        level_idx=_.level_idx.first(),
        vector_ids=_.vector_id.collect(),
    )
    levels = levels.distinct()

    match_level_filter = alt.selection_point(fields=["comparer", "level"])
    # Use names based off of https://github.com/altair-viz/altair/issues/2366
    vector_ids_filter = alt.selection_point(fields=["vector_ids"], name="vidf")

    legend_base = (
        alt.Chart(levels.to_pandas(), height=100)
        .mark_rect()
        .encode(
            x=alt.X(
                "level_idx:N", axis=alt.Axis(title="Level", labels=False, ticks=False)
            ),
            y=alt.Y(
                "comparer",
                title="Comparer",
            ),
            opacity=alt.condition(match_level_filter, alt.value(1), alt.value(0.4)),
            tooltip=["level"],
        )
    )
    legend_rects = legend_base.encode(
        color=alt.Color(
            "id",
            title="Comparer:Level",
            scale=alt.Scale(domain=color_map[0], range=color_map[1]),
            legend=None,
        ),
        tooltip=["level"],
    )
    legend_text = legend_base.mark_text(
        align="center",
        baseline="middle",
        color="white",
    ).encode(
        text="level",
    )
    legend = legend_rects + legend_text
    legend = legend.add_params(match_level_filter)
    legend = legend.add_params(vector_ids_filter)
    return legend



from __future__ import annotations

import functools
from typing import overload

import ibis
from ibis.expr import types as ir

from mismo._util import optional_import


@overload
def are_aliases(name1: ir.StringValue, name2: ir.StringValue) -> ir.BooleanValue: ...


@overload
def are_aliases(name1: str, name2: str) -> bool: ...


def are_aliases(name1, name2):
    """Determine if two names are nickname interchangeable.

    This is case-insensitive, and whitespace is stripped from both ends.
    The same name returns True.
    """
    is_string = False
    if isinstance(name1, str) and isinstance(name2, str):
        is_string = True
    if isinstance(name1, str):
        name1 = ibis.literal(name1)
    if isinstance(name2, str):
        name2 = ibis.literal(name2)
    result = _are_aliases(name1, name2)
    if is_string:
        return bool(result.execute())
    return result


@overload
def is_nickname_for(
    nickname: ir.StringValue, canonical: ir.StringValue
) -> ir.BooleanValue: ...


@overload
def is_nickname_for(nickname: str, canonical: str) -> bool: ...


def is_nickname_for(nickname, canonical):
    """Determine if a name is a nickname for another name.

    This is case-insensitive, and whitespace is stripped from both ends.
    The same name returns True.
    """
    is_string = False
    if isinstance(nickname, str) and isinstance(canonical, str):
        is_string = True
    if not isinstance(nickname, ir.StringValue):
        nickname = ibis.literal(nickname)
    if not isinstance(canonical, ir.StringValue):
        canonical = ibis.literal(canonical)
    result = _is_nickname_for(nickname, canonical)
    if is_string:
        return bool(result.execute())
    return result


def _is_nickname_for(
    nickname: ir.StringValue, canonical: ir.StringValue
) -> ir.BooleanValue:
    nickname = nickname.lower().strip()
    canonical = canonical.lower().strip()
    needle = ibis.struct({"canonical": canonical, "nickname": nickname})
    result = needle.isin(_nicknames_column()) | (canonical == nickname)
    return result


def _are_aliases(name1: ir.StringValue, name2: ir.StringValue) -> ir.BooleanValue:
    name1 = name1.lower().strip()
    name2 = name2.lower().strip()
    needle = ibis.struct({"name1": name1, "name2": name2})
    result = needle.isin(_aliases_column()) | (name1 == name2)
    return result


def _nicknames_pairs() -> set[tuple[str, str]]:
    with optional_import("nicknames"):
        from nicknames import NickNamer

    nn = NickNamer()
    pairs = set()
    for canonical, nicknames in nn._nickname_lookup.items():
        for nickname in nicknames:
            pairs.add((canonical, nickname))
    return pairs


def _aliases_pairs() -> set[tuple[str, str]]:
    pairs = set()
    for canonical, nickname in _nicknames_pairs():
        pairs.add((canonical, nickname))
        pairs.add((nickname, canonical))
    return pairs


@functools.cache
def _nicknames_column() -> ir.StructColumn:
    pairs = _nicknames_pairs()
    records = [{"canonical": c, "nickname": n} for c, n in pairs]
    t = ibis.memtable(
        {"pairs": records},
        schema={"pairs": "struct<canonical: string, nickname: string>"},
    )
    return t.pairs


@functools.cache
def _aliases_column() -> ir.StructColumn:
    pairs = _aliases_pairs()
    records = [{"name1": c, "name2": n} for c, n in pairs]
    t = ibis.memtable(
        {"pairs": records},
        schema={"pairs": "struct<name1: string, name2: string>"},
    )
    return t.pairs



from __future__ import annotations

import warnings

import ibis
import pytest

import mismo
from mismo.tests.util import assert_tables_equal


@pytest.fixture
def people_left(table_factory):
    """A simple test dataset representing people."""
    return table_factory(
        {
            "record_id": [0, 1, 2, 3, 4],
            "name": [
                "Alice Smith",
                "Bob Jones",
                "Charlie Brown",
                "Diana Prince",
                "Eve Adams",
            ],
            "age": [25, 30, 35, 28, 22],
            "city": ["New York", "Boston", "Chicago", "Seattle", "Denver"],
            "email": [
                "alice@email.com",
                "bob@email.com",
                "charlie@email.com",
                "diana@email.com",
                "eve@email.com",
            ],
        }
    )


@pytest.fixture
def people_right(table_factory):
    """Another simple test dataset representing people with some matches."""
    return table_factory(
        {
            "record_id": [10, 11, 12, 13, 14],
            "name": [
                "Alice Smith",
                "Robert Jones",
                "Charles Brown",
                "Wonder Woman",
                "Eve A",
            ],
            "age": [25, 30, 36, 28, 22],
            "city": ["New York", "Boston", "Chicago", "Seattle", "Denver"],
            "email": [
                "alice@email.com",
                "robert@email.com",
                "cb@email.com",
                "ww@email.com",
                "eve@email.com",
            ],
        }
    )


def test_or_linker_empty_conditions(table_factory, people_left, people_right):
    """Test OrLinker with no conditions returns empty linkage."""
    linker = mismo.OrLinker([])
    linkage = linker(people_left, people_right)
    expected = table_factory([], schema=linkage.links.schema())
    assert_tables_equal(expected, linkage.links)


def test_or_linker_single_condition(table_factory, people_left, people_right):
    """Test OrLinker with a single condition."""
    conditions = ["name"]
    linker = mismo.OrLinker(conditions)
    linkage = linker(people_left, people_right)

    actual = linkage.links.select("record_id_l", "record_id_r").order_by(
        "record_id_l", "record_id_r"
    )
    expected = table_factory(
        [
            (0, 10),  # Alice Smith
        ],
        schema=actual.schema(),
    )
    assert_tables_equal(expected, actual, order_by=["record_id_l", "record_id_r"])


@pytest.mark.parametrize(
    "condition_getter",
    [
        pytest.param(lambda: ("name", "email"), id="string_conditions"),
        pytest.param(lambda: (ibis._.name, ibis._.email), id="deferred_conditions"),
        pytest.param(
            lambda: (mismo.KeyLinker("name"), mismo.KeyLinker("email")),
            id="key_linker_conditions",
        ),
        pytest.param(
            lambda: (
                lambda left, right: left.name == right.name,
                lambda left, right: left.email == right.email,
            ),
            id="lambda_conditions",
        ),
        pytest.param(lambda: ("name", mismo.KeyLinker("email")), id="mixed_conditions"),
        # including False shouldn't affect the linkage
        pytest.param(
            lambda: ("name", mismo.KeyLinker("email"), False), id="with_false"
        ),
        pytest.param(
            lambda: (("name", "name"), mismo.KeyLinker("email")), id="tuple_conditions"
        ),
    ],
)
def test_or_linker_multiple_conditions(
    table_factory, people_left, people_right, condition_getter
):
    """Test OrLinker with multiple conditions (OR logic)."""
    conditions = condition_getter()
    linker = mismo.OrLinker(conditions)
    linkage = linker(people_left, people_right)

    actual = linkage.links.select("record_id_l", "record_id_r").order_by(
        "record_id_l", "record_id_r"
    )
    expected = table_factory(
        [
            (0, 10),  # Alice Smith - matches both name and email
            # even though Alice Smith matches multiple conditions, it doesn't appear
            # twice, because we removed overlapping conditions.
            (4, 14),  # Eve Adams / Eve A - matches email
        ],
        schema=actual.schema(),
    )
    assert_tables_equal(expected, actual, order_by=["record_id_l", "record_id_r"])


def test_or_linker_with_named_conditions(table_factory, people_left, people_right):
    """Test OrLinker with named conditions (dict input)."""
    conditions = {
        "name_match": "name",
        "email_match": "email",
    }
    linker = mismo.OrLinker(conditions)
    linkage = linker(people_left, people_right)

    # Check that we can access the join_conditions by name
    assert "name_match" in linker.join_conditions
    assert "email_match" in linker.join_conditions

    # Should match on name: Alice Smith (1,10), Charles Brown (3,12)
    # Should match on email: Alice Smith (1,10), Eve Adams (5,14)
    actual = linkage.links.select("record_id_l", "record_id_r").order_by(
        "record_id_l", "record_id_r"
    )
    expected = table_factory(
        [
            (0, 10),  # Alice Smith - matches both name and email
            (4, 14),  # Eve Adams / Eve A - matches email
        ],
        schema=actual.schema(),
    )
    assert_tables_equal(expected, actual, order_by=["record_id_l", "record_id_r"])


def test_or_linker_deduplication(table_factory):
    """Test OrLinker on deduplication (same table)."""
    people = table_factory(
        {
            "record_id": [1, 2, 3, 4],
            "name": ["Alice Smith", "Alice Smith", "Bob Jones", "Charlie Brown"],
            "age": [25, 25, 30, 35],
            "email": [
                "alice@email.com",
                "alice2@email.com",
                "bob@email.com",
                "charlie@email.com",
            ],
        }
    )

    conditions = ["name", "age"]
    linker = mismo.OrLinker(conditions)
    linkage = linker(people, people)

    # For deduplication, we expect record_id_l < record_id_r
    # Should match on name: Alice Smith records (1,2)
    # Should match on age: Alice Smith records (1,2) - already covered by name
    actual = linkage.links.select("record_id_l", "record_id_r").order_by(
        "record_id_l", "record_id_r"
    )
    expected = table_factory(
        [
            (1, 2),  # Alice Smith records - match on both name and age
        ],
        schema=actual.schema(),
    )
    assert_tables_equal(expected, actual, order_by=["record_id_l", "record_id_r"])


def test_or_linker_on_slow_parameter(people_left, people_right):
    """Test OrLinker with on_slow parameter."""
    # This should not raise an error even with potentially slow conditions
    conditions = [
        lambda left, right: left.name.levenshtein(right.name) < 2,  # Potentially slow
    ]

    # Test with ignore
    linker = mismo.OrLinker(conditions, on_slow="ignore")
    linkage = linker(people_left, people_right)
    assert linkage is not None

    # Test with warn (should warn but not error)
    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        linker = mismo.OrLinker(conditions, on_slow="warn")
        linkage = linker(people_left, people_right)
        assert isinstance(linkage, mismo.Linkage)

    with pytest.raises(mismo.exceptions.SlowJoinError):
        # Test with error (should raise SlowJoinError)
        linker = mismo.OrLinker(conditions, on_slow="error")
        linker(people_left, people_right)


def test_or_linker_properties():
    """Test OrLinker properties and methods."""
    conditions = {
        "name_match": "name",
        "email_match": "email",
    }
    linker = mismo.OrLinker(conditions)

    # Test join_conditions property
    assert len(linker.join_conditions) == 2
    assert "name_match" in linker.join_conditions
    assert "email_match" in linker.join_conditions

    # Test that each condition has __join_condition__ method
    for condition in linker.join_conditions.values():
        assert hasattr(condition, "__join_condition__")



from __future__ import annotations

import ibis
import pytest

from mismo.lib import email


@pytest.mark.parametrize(
    "input, exp",
    [
        pytest.param(
            " Bob.Jones@gmail.com",
            "bob.jones@gmail.com",
            id="bobjones",
        ),
        pytest.param("", None, id="empty"),
        pytest.param("  ", None, id="whitespace"),
        pytest.param(None, None, id="null"),
        pytest.param("@", None, id="at"),
        pytest.param(" @  ", None, id="empty_at_empty"),
        pytest.param("anne@", None, id="no_domain"),
        pytest.param("@gmail.com", None, id="no_user"),
        pytest.param("A@B ", "a@b", id="short"),
    ],
)
def test_clean_email(input, exp):
    result = email.clean_email(ibis.literal(input, type="string")).execute()
    assert result == exp


@pytest.mark.parametrize(
    "input, expfull, expuser, expdomain",
    [
        pytest.param(
            "bob.jones@gmail.com",
            "bob.jones@gmail.com",
            "bob.jones",
            "gmail.com",
            id="bobjones",
        ),
        pytest.param("", "", None, None, id="empty"),
        pytest.param(None, None, None, None, id="null"),
    ],
)
def test_ParsedEmail(input, expfull, expuser, expdomain):
    parsed = email.ParsedEmail(ibis.literal(input, str))
    assert parsed.full.execute() == expfull
    assert parsed.user.execute() == expuser
    assert parsed.domain.execute() == expdomain

    assert parsed.as_struct().execute() == {
        "full": expfull,
        "user": expuser,
        "domain": expdomain,
    }


@pytest.mark.parametrize(
    "a, b, level_str",
    [
        pytest.param(
            " Bob.Jones@gmail.com", "bobjones@gmail.com", "FULL_EXACT", id="full_exact"
        ),
        pytest.param(
            " Bob.Janes@gmail.com", "bobjones@gmail.com", "FULL_NEAR", id="full_near"
        ),
        pytest.param(
            " Rob.Janes@gmail.com", "bobjones@gmail.com", "ELSE", id="two_different"
        ),
        pytest.param(
            " Bob.Jones@gmail.com", "bobjones@yahoo.com", "USER_EXACT", id="user_exact"
        ),
        pytest.param(
            " Bob.Janes@gmail.com", "bobjones@yahoo.com", "USER_NEAR", id="user_near"
        ),
    ],
)
def test_match_level(a, b, level_str):
    ml = email.match_level(ibis.literal(a), ibis.literal(b))
    assert ml.as_string().execute() == level_str
    assert ml.as_integer().execute() == email.EmailMatchLevel[level_str]



from __future__ import annotations

from typing import Literal, Protocol, runtime_checkable

import ibis

from mismo.linkage import _linkage


@runtime_checkable
class Linker(Protocol):
    """
    A Protocol that takes two tables of records and produces a [Linkage][mismo.Linkage].
    """

    def __call__(self, left: ibis.Table, right: ibis.Table) -> _linkage.Linkage:
        """Given two tables, return a Linkage."""
        raise NotImplementedError


def infer_task(
    *, task: Literal["dedupe", "link"] | None, left: ibis.Table, right: ibis.Table
) -> Literal["dedupe", "link"]:
    if task is not None:
        return task
    if left.equals(right):
        return "dedupe"
    return "link"



from __future__ import annotations

from collections.abc import Iterable
from typing import Callable

import ibis
from ibis import _
from ibis.expr import datatypes as dt
from ibis.expr import types as ir

from mismo import _util


def array_combinations(left: ir.ArrayValue, right: ir.ArrayValue) -> ir.ArrayValue:
    """Generate all combinations of elements from two arrays.

    This is the cartesian product of the two arrays.

    Parameters
    ----------
    left
        The first array.
    right
        The second array.

    Returns
    -------
    combinations
        An `array<struct<l: T, r: U>>` where `T` is the type of the
        elements in `array1` and `U` is the type of the elements in `array2`.
    """
    return left.map(
        lambda le: right.map(lambda r: ibis.struct(dict(l=le, r=r)))
    ).flatten()


def array_combinations_first_n(
    left: ir.ArrayValue,
    right: ir.ArrayValue,
    *,
    n_left: int,
    n_right: int,
) -> list[tuple[ir.Value, ir.Value]]:
    """Generate the combinations from the first N elements of two arrays.

    This is a static version of `array_combinations`.
    That function generates ALL combinations of the two arrays,
    but since that number is dynamic, it can be hard to work with.
    This function is useful for if you have a small number of
    combinations you want to generate,
    eg you want to compare all the email addresses for
    one person to all the email addresses for another person.

    Parameters
    ----------
    left
        The first array.
    right
        The second array.
    n_left
        The number of elements to take from the first array.
    n_right
        The number of elements to take from the second array.

    Returns
    -------
    combinations
        A list of (left, right) tuples.
    """
    if n_left < 0 or n_right < 0:
        raise ValueError("n_left and n_right must be non-negative.")
    results = []
    for i in range(n_left):
        for j in range(n_right):
            results.append((left[i], right[j]))
    return results


def array_filter_isin_other(
    t: ir.Table,
    array: ir.ArrayColumn | str,
    other: ir.Column,
    *,
    result_format: str = "{name}_filtered",
) -> ir.Table:
    """
    Equivalent to `t.mutate(result_name=t[array].filter(lambda x: x.isin(other)))`

    We can't have subqueries in the filter lambda (need this to avoid
    [https://stackoverflow.com/questions/77559936/how-to-implementlist-filterarray-elem-elem-in-column-in-other-table]()))

    See [issues/32](https://github.com/NickCrews/mismo/issues/32) for more info.

    Parameters
    ----------
    t :
        The table containing the array column.
    array :
        A reference to the array column.
    other :
        The column to filter against.
    result_format :
        The format string to use for the result column name. The format string
        should have a single placeholder, `{name}`, which will be replaced with
        the name of the array column.

    Returns
    -------
    ir.Table
        The table with a new column named following `result_format` with the
        filtered array.
    """  # noqa E501
    array_val = _util.bind_one(t, array)
    t = t.mutate(__array=array_val, __id=ibis.row_number())
    temp = t.select("__id", __unnested=_.__array.unnest())
    # When we re-.collect() items below, the order matters,
    # but the .filter() can mess up the order, so we need to
    # add a sortable key before the filter
    filtered = temp.mutate(elem_id=ibis.row_number()).filter(
        temp.__unnested.isin(other) | temp.__unnested.isnull()
    )
    re_agged = filtered.group_by("__id").agg(
        __filtered=filtered.__unnested.collect(include_null=True, order_by="elem_id")
    )
    re_joined = t.left_join(re_agged, "__id").drop("__id", "__id_right")
    # when we unnested, both [] and NULL rows disappeared.
    # Once we join back, they all come back as NULL,
    # so we need to adjust some of them back to [].
    re_joined = re_joined.mutate(
        __filtered=(_.__array.notnull() & _.__filtered.isnull()).ifelse(
            [], _.__filtered
        )
    ).drop("__array")
    result_name = result_format.format(name=array_val.get_name())
    return re_joined.rename({result_name: "__filtered"})


def _list_select(x: ir.ArrayValue, indexes: ir.ArrayValue) -> ir.ArrayValue:
    """Selects elements from a list by index."""

    t = x.type()
    # if t.is_array():
    #     raise ValueError(f"Expected an array, got {t}")

    @ibis.udf.scalar.builtin(name="list_select", signature=((t, "array<int>"), t))
    def f(array, idxs): ...

    return f(x, indexes)


@ibis.udf.scalar.builtin(
    name="list_grade_up",
    # signature=(("array<float64>",), "array<int64>"),
)
def _list_grade_up(x) -> dt.Array(value_type=dt.Int64()):
    """Works like sort, but returns the indexes instead of the actual values."""


def array_shuffle(a: ir.ArrayValue) -> ir.ArrayValue:
    """Shuffle an array."""
    idxs = a.map(lambda x: ibis.random())
    return _list_select(a, _list_grade_up(idxs))


def array_choice(a: ir.ArrayValue, n: int) -> ir.ArrayValue:
    """Randomly select `n` elements from an array."""
    return array_shuffle(a)[n:]


def array_sort(
    arr: ir.ArrayValue,
    /,
    *,
    key: ibis.Deferred
    | Callable[[ir.Value], ir.Value | Iterable[ir.Value]]
    | Iterable[ibis.Deferred]
    | None = None,
) -> ir.ArrayValue:
    """Sort an array, optionally using a key.

    The builtin ArrayValue.sort() method doesn't support a key.
    This is a workaround for that.

    See https://github.com/duckdb/duckdb/discussions/10417
    """
    if key is None:
        return arr.sort()

    def resolve_key(elem: ir.Value, k):
        if isinstance(k, ir.Value):
            return k
        if isinstance(k, ibis.Deferred):
            return k.resolve(elem)
        if callable(k):
            return resolve_key(elem, k(elem))
        # assume iterable of values
        return ibis.struct(
            {f"f{i}": resolve_key(elem, subkey) for i, subkey in enumerate(k)}
        )

    keys = arr.map(lambda x: resolve_key(x, key))
    return _list_select(arr, _list_grade_up(keys))



from __future__ import annotations

import ibis
import pytest

import mismo
from mismo.linker._lsh import p_blocked

sets = [
    # set(),
    {0},
    # {0, 1},
    {0, 1, 2},
    # {0, 1, 2, 3},
]


@pytest.mark.skip(reason="MinhashLshBlocker isn't implemented correctly")
@pytest.mark.parametrize("a", sets, ids=str)
@pytest.mark.parametrize("b", sets, ids=str)
@pytest.mark.parametrize("band_size", [5])
@pytest.mark.parametrize("n_bands", [10])
def test_minhash_lsh_blocker(table_factory, a, b, band_size: int, n_bands: int):
    left = table_factory({"terms": [a] * 100}, schema={"terms": "array<int>"})
    right = table_factory({"terms": [b] * 100}, schema={"terms": "array<int>"})
    left = left.mutate(record_id=ibis.row_number())
    right = right.mutate(record_id=ibis.row_number())
    blocker = mismo.linkage.MinhashLshBlocker(
        terms_column="terms", band_size=band_size, n_bands=n_bands
    )
    p_expected = p_blocked(_jaccard(a, b), band_size=band_size, n_bands=n_bands)
    n_expected = (100 * 100) * p_expected
    blocked = blocker(left, right).execute()
    assert len(blocked) == pytest.approx(n_expected, rel=0.1)


def _jaccard(a, b):
    if not a and not b:
        return 0
    return len(a & b) / len(a | b)



from __future__ import annotations

from collections.abc import Iterable, Mapping
import functools
from typing import Generic, Literal, TypeVar

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo import _util
from mismo.types._wrapper import TableWrapper

T = TypeVar("T", bound=ir.Column)


class StatsTable(TableWrapper, Generic[T]):
    value: T
    count: ir.IntegerColumn
    frequency: ir.FloatingColumn

    @classmethod
    def from_column(cls, column: T) -> StatsTable[T]:
        t = column.name("value").as_table()
        n_total = t.count().as_scalar()
        t = t.aggregate(
            count=_.count().cast("uint32"),
            frequency=_.count() / n_total,
            by="value",
        )
        return cls(t)


class ColumnStats:
    """Represents statistics for a column."""

    def __init__(self, column: ir.Column):
        self.column = column
        self._stats_table = StatsTable.from_column(column)

    @property
    def name(self) -> str:
        """The name of the term."""
        return self.column.get_name()

    @functools.cached_property
    def stats_table(self) -> StatsTable:
        """The stats table for this column."""
        return StatsTable(self._stats_table.cache())

    def add_frequencies(
        self,
        table: ibis.Table,
        *,
        column: str | ibis.Deferred | ibis.Value | None = None,
        name_as: str | None = None,
        default: Literal["1/N"] | int | float | ibis.Scalar = "1/N",
    ) -> ibis.Table:
        """Add frequency columns to the given table."""
        if name_as is None:
            name_as = f"frequency_{self.name}"
        if column is None:
            column = self.name

        default_resolved: ibis.Scalar
        if default == "1/N" or default == "1/n":
            n_total = table.count().as_scalar()
            default_resolved = (1 / n_total).cast("float64")
        elif isinstance(default, ibis.Scalar):
            default_resolved = default.cast("float64")  # ty:ignore[invalid-assignment]
        else:
            default_resolved = ibis.literal(default, "float64")
        table_column = _util.bind_one(table, column)

        unique_name = _util.unique_name("join_key")
        # TODO: this could be factored out into a join_lookup() function
        stats_raw = self.stats_table.select(
            _.value.name(unique_name), _.frequency.name(name_as)
        )
        filler = (
            table_column.name(unique_name)
            .as_table()
            .distinct()
            .anti_join(stats_raw, unique_name)
            .mutate(default_resolved.name(name_as))
        )
        stats = ibis.union(stats_raw, filler)
        assert stats.columns == (unique_name, name_as)
        table = table.left_join(stats, table_column == stats[unique_name]).drop(
            unique_name
        )
        return table


class TermFrequencyModel:
    columns: dict[str, ir.Column]

    def __init__(
        self,
        columns: ibis.Table
        | ibis.Column
        | Mapping[str, ibis.Column]
        | Iterable[ibis.Column],
        /,
    ):
        if isinstance(columns, Mapping):
            self.columns = dict(columns)
        elif isinstance(columns, ibis.Table):
            self.columns = {col: columns[col] for col in columns.columns}
        elif isinstance(columns, ibis.Column):
            self.columns = {columns.get_name(): columns}
        else:
            self.columns = {col.get_name(): col for col in columns}

    @property
    def column_stats(self) -> dict[str, ColumnStats]:
        return {
            name: ColumnStats(column.name(name))
            for name, column in self.columns.items()
        }

    def add_term_frequencies(
        self,
        table: ibis.Table,
        *,
        columns: dict[str, str | ibis.Deferred | ibis.Column] | None = None,
        name_as: str | None = None,
        default: Literal[0, "1/N"] = "1/N",
    ) -> ibis.Table:
        """Add frequency columns for all terms in the model to the table."""
        if name_as is None:
            name_as = "term_frequencies"
        freq_names = []
        for column_name, term in self.column_stats.items():
            unique_name = _util.unique_name("freq")
            column = None if columns is None else columns.get(column_name)
            table = term.add_frequencies(
                table, column=column, name_as=unique_name, default=default
            )
            freq_names.append(unique_name)
        table = table.mutate(
            _product(*[table[name] for name in freq_names]).name(name_as)
        ).drop(*freq_names)
        return table

    def __repr__(self):
        return f"{self.__class__.__name__}({list(self.columns.keys())})"


def _product(*args: ir.Column) -> ir.Column:
    result = args[0]
    for arg in args[1:]:
        result = result * arg
    return result



"""Compare module.

This module contains functions and classes for comparing pairs of records.
"""

from __future__ import annotations

from mismo.compare._comparer import PComparer as PComparer
from mismo.compare._match_level import LevelComparer as LevelComparer
from mismo.compare._match_level import MatchLevel as MatchLevel
from mismo.compare._plot import compared_dashboard as compared_dashboard



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Mapping

from ibis.expr import types as ir

from mismo._datasets import Datasets

if TYPE_CHECKING:
    import solara


def cluster_dashboard(
    ds: Datasets | ir.Table | Iterable[ir.Table] | Mapping[str, ir.Table],
    links: ir.Table,
) -> solara.Column:
    """A Solara component for that shows a cluster of records and links.

    This shows *ALL* the supplied records and links, so be careful,
    you probably want to filter them down first.
    You can use the `clusters_dashboard` component for that.

    This is like `cytoscape_widget`, but with a status bar that shows
    information about the selected node or edge.

    Parameters
    ----------
    ds :
        Table(s) of records with at least the column `record_id`.
    links :
        A table of edges with at least columns
        (record_id_l, record_id_r) and optionally other columns.
        The column `width` is used to set the width of the edges.
        If not given, it is determined from the column `odds`, if
        present, or set to 5 otherwise.
        The column `opacity` is used to set the opacity of the edges.
        If not given, it is set to 0.5.
    """
    from mismo.cluster._dashboard_internal import cluster_dashboard

    return cluster_dashboard(ds, links)


def clusters_dashboard(
    tables: Datasets | ir.Table | Iterable[ir.Table] | Mapping[str, ir.Table],
    links: ir.Table,
) -> solara.Column:
    """Make a dashboard for exploring different clusters of records.

    Pass the entire dataset and the links between records,
    and use this to filter down to a particular cluster.
    """
    from mismo.cluster._dashboard_internal import clusters_dashboard

    return clusters_dashboard(tables, links)



from __future__ import annotations

from collections.abc import Mapping
import dataclasses
import math
from typing import Callable, Literal, TypedDict

import ibis
from ibis.expr import types as ir

from mismo import _util, linkage, types
from mismo._counts_table import KeyCountsTable, PairCountsTable
from mismo.linker._key_linker import KeyLinker


def distance_km(
    *,
    lat1: ir.FloatingValue,
    lon1: ir.FloatingValue,
    lat2: ir.FloatingValue,
    lon2: ir.FloatingValue,
) -> ir.FloatingValue:
    """The distance between two points on the Earth's surface, in kilometers.

    Parameters
    ----------
    lat1:
        The latitude of the first point.
    lon1:
        The longitude of the first point.
    lat2:
        The latitude of the second point.
    lon2:
        The longitude of the second point.

    Returns
    -------
    distance:
        The distance between the two points, in kilometers.
    """

    def radians(degrees):
        scaling = math.pi / 180
        return degrees * scaling

    lat1 = radians(lat1)
    lon1 = radians(lon1)
    lat2 = radians(lat2)
    lon2 = radians(lon2)

    def haversine(theta: ir.FloatingValue) -> ir.FloatingValue:
        return (theta / 2).sin() ** 2

    a = haversine(lat2 - lat1) + lat1.cos() * lat2.cos() * haversine(lon2 - lon1)
    R_earth = 6371.0  # km
    # parens for optimization, to make sure the division happens on the python side
    return (R_earth * 2) * a.sqrt().asin()


Coordinates = Mapping[Literal["lat", "lon"], ir.FloatingColumn]
CoordinatesMapping = Mapping[Literal["lat", "lon"], _util.IntoValue]


class CoordinatesDict(TypedDict):
    lat: ir.FloatingColumn
    lon: ir.FloatingColumn


IntoCoordinates = (
    str
    | ibis.Deferred
    | Coordinates
    | CoordinatesMapping
    | Callable[[ir.Table], "IntoCoordinates"]
)


def default_resolver() -> CoordinatesMapping:
    return {"lat": "lat", "lon": "lon"}


def get_coordinate_pair(
    table: ir.Table,
    mapping: IntoCoordinates,
) -> CoordinatesDict:
    """Get the coordinates from a table.

    Parameters
    ----------
    table
        The table to get the coordinates from.
    mapping
        A mapping of column names to use for the coordinates,
        or a function that takes a table and returns the coordinate pair.

    Returns
    -------
    coordinates
        The coordinates.
    """
    if callable(mapping):
        called = mapping(table)
        return get_coordinate_pair(table, called)

    if isinstance(mapping, str) or isinstance(mapping, ibis.Deferred):
        coords_col = _util.bind_one(table, mapping)
        lat: ir.FloatingColumn = coords_col["lat"]  # ty:ignore[not-subscriptable]
        lon: ir.FloatingColumn = coords_col["lon"]  # ty:ignore[not-subscriptable]
    else:
        lat: ir.FloatingColumn = _util.bind_one(table, mapping["lat"])  # ty:ignore[invalid-assignment]
        lon: ir.FloatingColumn = _util.bind_one(table, mapping["lon"])  # ty:ignore[invalid-assignment]
    return CoordinatesDict(lat=lat, lon=lon)


@dataclasses.dataclass(frozen=True)
class CoordinateLinker:
    """Links two locations together if they are within a certain distance.

    This isn't precise, and can include pairs that are actually up to about 2x
    larger than the given threshold.
    This is because we use a simple grid to bin the coordinates, so
    1. This isn't accurate near the poles, and
    2. This isn't accurate near the international date line (longitude 180/-180).
    3. If two coords fall within opposite corners of the same grid cell, they
        will be blocked together even if they are further apart than the
        precision, due to the diagonal distance being longer than the horizontal
        or vertical distance.

    Examples
    --------
    >>> import ibis
    >>> from mismo.lib.geo import CoordinateLinker
    >>> ibis.options.interactive = True
    >>> conn = ibis.duckdb.connect()
    >>> left = conn.create_table(
    ...     "left",
    ...     [
    ...         {
    ...             "record_id": 0,
    ...             "latlon": {"lat": 61.1547800, "lon": -150.0677490},
    ...         }
    ...     ],
    ... )
    >>> right = conn.create_table(
    ...     "right",
    ...     [
    ...         {
    ...             "record_id": 4,
    ...             "latitude": 61.1582056,
    ...             "longitude": -150.0584552,
    ...         },
    ...         {
    ...             "record_id": 5,
    ...             "latitude": 61.1582056,
    ...             "longitude": 0,
    ...         },
    ...         {
    ...             "record_id": 6,
    ...             "latitude": 61.1547800,
    ...             "longitude": -150,
    ...         },
    ...     ],
    ... )
    >>> linker = CoordinateLinker(
    ...     distance_km=1,
    ...     left_resolver="latlon",
    ...     right_resolver={"lat": "latitude", "lon": "longitude"},
    ... )
    >>> linker(left, right).links
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━┓
    ┃ record_id_l ┃ latlon_l                      ┃ record_id_r ┃ latitude_r ┃ longitude_r ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━┩
    │ int64       │ struct<lat: float64, lon:     │ int64       │ float64    │ float64     │
    │             │ float64>                      │             │            │             │
    ├─────────────┼───────────────────────────────┼─────────────┼────────────┼─────────────┤
    │             │ {                             │             │            │             │
    │           0 │     'lat': 61.15478,          │           4 │  61.158206 │ -150.058455 │
    │             │     'lon': -150.067749        │             │            │             │
    │             │ }                             │             │            │             │
    └─────────────┴───────────────────────────────┴─────────────┴────────────┴─────────────┘
    """  # noqa: E501

    distance_km: float | int
    """
    The (approx) max distance in kilometers that two coords will be blocked together.
    """
    max_pairs: int | None = None
    """The maximum number of pairs that any single block of coordinates can contain.
    
    eg if you have 1000 records all with the same coordinates, this would
    naively result in ~(1000 * 1000) / 2 = 500_000 pairs.
    If we set max_pairs to less than this, this group of records will be skipped.
    """
    left_resolver: IntoCoordinates = dataclasses.field(default_factory=default_resolver)
    """A specification of how to get the lat and lon values from the left table.
    
    Can be:
    - A `str` or `ibis.Deferred`, which are assumed to point to a column
      containing a struct with `lat` and `lon` fields.
    - A `Mapping` of `{"lat": ..., "lon": ...}` where the values are
      column names or `ibis.Deferred` expressions.
    - A callable that takes a table and returns one of the above.
    """
    right_resolver: IntoCoordinates = dataclasses.field(
        default_factory=default_resolver
    )
    """See `left_resolver`, but for the right table."""

    def hash_coord(
        self, coord: CoordinatesDict, /
    ) -> tuple[ir.IntegerValue, ir.IntegerValue]:
        lat, lon = coord["lat"], coord["lon"]
        # We have to use a grid size of ~3x the precision to avoid
        # two points falling right on either side of a grid cell boundary
        grid_size = self.distance_km * 3
        lat_key, lon_key = _bin_lat_lon(lat, lon, grid_size)
        return lat_key, lon_key

    def hash_left(self, t: ir.Table, /) -> tuple[ir.IntegerValue, ir.IntegerValue]:
        coords = get_coordinate_pair(t, self.left_resolver)
        return self.hash_coord(coords)

    def hash_right(self, t: ir.Table, /) -> tuple[ir.IntegerValue, ir.IntegerValue]:
        coords = get_coordinate_pair(t, self.right_resolver)
        return self.hash_coord(coords)

    @property
    def _key_linker(self) -> KeyLinker:
        import mismo

        def lat_key_left(t: ir.Table) -> ir.IntegerValue:
            return self.hash_left(t)[0]

        def lon_key_left(t: ir.Table) -> ir.IntegerValue:
            return self.hash_left(t)[1]

        def lat_key_right(t: ir.Table) -> ir.IntegerValue:
            return self.hash_right(t)[0]

        def lon_key_right(t: ir.Table) -> ir.IntegerValue:
            return self.hash_right(t)[1]

        return mismo.KeyLinker(
            [(lat_key_left, lat_key_right), (lon_key_left, lon_key_right)],
            max_pairs=self.max_pairs,
        )

    def __join_condition__(self, left: ir.Table, right: ir.Table) -> ir.BooleanValue:
        return self._key_linker.__join_condition__(left, right)

    def __call__(self, left: ir.Table, right: ir.Table) -> linkage.Linkage:
        links = types.LinksTable.from_join_condition(left, right, self)
        return linkage.Linkage(left=left, right=right, links=links)

    def key_counts_left(self, left: ibis.Table, /) -> KeyCountsTable:
        return self._key_linker.key_counts_left(left)

    def key_counts_right(self, right: ibis.Table, /) -> KeyCountsTable:
        return self._key_linker.key_counts_right(right)

    def pair_counts(self, left: ibis.Table, right: ibis.Table) -> PairCountsTable:
        return self._key_linker.pair_counts(left, right)


def _bin_lat_lon(
    lat: ir.FloatingValue, lon: ir.FloatingValue, grid_size_km: float | int
) -> tuple[ir.IntegerValue, ir.IntegerValue]:
    """Bin a latitude or longitude to a grid of a given precision.

    Say you have two coordinates, (lat1, lon1) and (lat2, lon2), and you
    want to see if they are within, say 15km of each other. You can bin
    them both to a grid of 15km precision and compare the resulting
    hashed values. If they are the same, the coordinates are within 15km
    of each other.
    """
    km_per_lat, km_per_lon = _km_per_degree(lat)
    step_size_lat = grid_size_km / km_per_lat
    step_size_lon = grid_size_km / km_per_lon
    if ibis.__version__ >= "10.0.0":
        # See https://github.com/ibis-project/ibis/pull/10353
        lat_hash = lat // step_size_lat
        lon_hash = lon // step_size_lon
    else:
        lat_hash = (lat / step_size_lat).floor().cast(int)
        lon_hash = (lon / step_size_lon).floor().cast(int)
    both_null = lat.isnull() & lon.isnull()
    lat_result = both_null.ifelse(ibis.null(), lat_hash)
    lon_result = both_null.ifelse(ibis.null(), lon_hash)
    return lat_result, lon_result


def _km_per_degree(lat: ir.FloatingValue) -> tuple[float, ir.FloatingValue]:
    # Radius of the Earth in kilometers
    R = 6371.0
    lat_rad = lat * (math.pi / 180)
    # This is a constant value at any given latitude.
    km_per_lat = (math.pi * R) / 180
    # This varies based on the latitude.
    # Parens for optimization, to make sure the division happens on the python side.
    km_per_lon = lat_rad.cos() * ((math.pi * R) / 180)
    return km_per_lat, km_per_lon



from __future__ import annotations

import ibis
from ibis.expr import types as ir

from mismo import _util


def norm_whitespace(texts: ir.StringValue) -> ir.StringValue:
    """
    Strip leading/trailing whitespace, replace multiple whitespace with a single space.
    """
    texts = _util.ensure_val(texts, "string")
    return texts.strip().re_replace(r"\s+", " ")


def strip_accents(s: ir.StringValue) -> ir.StringValue:
    """Remove accents, such as é -> e. Only works with duckdb.

    Parameters
    ----------
    s
        The string to strip

    Returns
    -------
    ir.StringValue
        The string with non-ascii characters replaced and/or removed.

    Examples
    --------
    >>> import ibis
    >>> strip_accents(ibis.literal("müller")).execute()
    'muller'
    >>> strip_accents(ibis.literal("François")).execute()
    'Francois'
    >>> strip_accents(ibis.literal("Øslo")).execute()  # Ø is not an accent
    'Øslo'
    >>> strip_accents(ibis.literal("æ")).execute()  # neither is this
    'æ'
    >>> strip_accents(ibis.literal("ɑɽⱤoW")).execute()  # neither is this
    'ɑɽⱤoW'
    """
    # There is a python lib unidecode, but I don't want us to take the performance hit.
    # polars also appears to implement this.
    # But there is no pyarrow implementation :(
    is_duckdb = (
        isinstance(s, ibis.Deferred)
        or s._find_backend(use_default=True).name == "duckdb"
    )
    if not is_duckdb:
        raise NotImplementedError("remove_accents not implemented for non-duckdb")
    s = _strip_accents(s)
    return s


@ibis.udf.scalar.builtin(name="strip_accents", signature=(("string",), "string"))
def _strip_accents(s: ir.StringValue) -> ir.StringValue: ...



from __future__ import annotations

from mismo.linker._basic import EmptyLinker as EmptyLinker
from mismo.linker._basic import FullLinker as FullLinker
from mismo.linker._common import Linker as Linker
from mismo.linker._id_linker import IDLinker as IDLinker
from mismo.linker._join_linker import JoinLinker as JoinLinker
from mismo.linker._key_linker import KeyLinker as KeyLinker
from mismo.linker._or_linker import OrLinker as OrLinker
from mismo.linker._unnest import UnnestLinker as UnnestLinker



from __future__ import annotations

import pytest

from mismo._factorizer import Factorizer
from mismo.tests.util import get_clusters


@pytest.mark.parametrize(
    "values, values_type",
    [
        pytest.param(
            [
                ("a", 0),
                ("a", 0),
                ("b", 1),
                ("b", 1),
                ("c", 2),
                (None, None),
            ],
            "string",
            id="simple",
        ),
        pytest.param(
            [],
            "string",
            id="empty_string",
        ),
        pytest.param(
            [],
            "int",
            id="empty_int",
        ),
        pytest.param(
            [(4, 4), (3, 3), (2, 2), (None, None)],
            "int64",
            id="int64",
        ),
        pytest.param(
            [(4, 2), (3, 1), (2, 0), (None, None)],
            "uint64",
            id="uint64",
        ),
        pytest.param(
            [
                (1.0, 0),
                (3.5, 1),
                (3.5, 1),
                (None, None),
            ],
            "float",
            id="float",
        ),
    ],
)
def test_factorizer(table_factory, values, values_type):
    data = {
        "id": range(len(values)),
        "values": [val for val, _code in values],
        "expected_codes": [code for _val, code in values],
    }
    schema = {
        "id": "int64",
        "values": values_type,
        "expected_codes": "int64",
    }
    t = table_factory(data).cast(schema)
    f = Factorizer(t, "values")

    e = f.encode()
    assert set(e.columns) == {"id", "values", "expected_codes"}
    assert_equal(e, "values", "expected_codes")

    e = f.encode(dst="encoded")
    assert set(e.columns) == {"id", "values", "expected_codes", "encoded"}
    assert_equal(e, "encoded", "expected_codes")

    restored = f.decode(e, src="encoded")
    assert set(restored.columns) == {"id", "values", "expected_codes", "encoded"}
    assert_equal(restored, "encoded", "values")

    restored = f.decode(e, src="encoded", dst="decoded")
    assert set(restored.columns) == {
        "id",
        "values",
        "expected_codes",
        "encoded",
        "decoded",
    }
    assert_equal(restored, "decoded", "values")

    t2 = t.mutate(values2=t.values)
    e = f.encode(t2)
    assert set(e.columns) == {"id", "values", "expected_codes", "values2"}
    assert_equal(e, "values", "expected_codes")

    e = f.encode(t2, src="values2", dst="codes2")
    assert set(e.columns) == {"id", "values", "expected_codes", "values2", "codes2"}
    assert_equal(e, "values", "values2")
    assert_equal(e, "codes2", "expected_codes")


def assert_equal(t, x, y):
    x_clusters = get_clusters(t[x], label=t.id)
    y_clusters = get_clusters(t[y], label=t.id)
    assert x_clusters == y_clusters



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, cast

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo import _common, _typing, _util
from mismo.types._links_table import LinksTable
from mismo.types._wrapper import TableWrapper

if TYPE_CHECKING:
    import altair as alt


class LinkedTable(TableWrapper):
    """A table of records that are linked to another table.

    Each record here can be linked to 0-N records in the `other_`.

    This acts just like an Ibis Table, but it has a few extra attributes
    and methods that make it more ergonomic to work with,
    eg to add data from the linked table.
    """

    def __init__(self, table: ibis.Table, *, other: ibis.Table, links: ibis.Table):
        _common.check_tables_and_links(table, other, links)

        super().__init__(table)
        object.__setattr__(self, "_other_raw", other)
        object.__setattr__(self, "_links_raw", links)

    @property
    def other_(self) -> _typing.Self:
        """The other table that this table is linked to.

        Trailing underscore to avoid name conflicts with column names."""
        # Need to make sure that we swap the _l and _r suffixes in the links table
        return self.__class__(
            self._other_raw,
            other=self,
            links=LinksTable._swap_perspective(self._links_raw),
        )

    @property
    def links_(self) -> LinksTable:
        """The table of links between this table and `other`.

        Trailing underscore to avoid name conflicts with column names."""
        return LinksTable(self._links_raw, left=self, right=self._other_raw)

    def with_linked_values(
        self,
        *values: ibis.Deferred | Callable[[ibis.Table], ir.Value] | None,
        default_agg: ibis.Deferred = _.collect(),
        **named_values: ibis.Deferred | Callable[[ibis.Table], ir.Value] | None,
    ) -> _typing.Self:
        """
        This table, with values from the linked records.

        See the examples below, this is easy to understand with examples,
        but hard to describe.

        Parameters
        ----------
        values
            unnamed values
        named_values
            named values

        Returns
        -------
        A new LinkedTable with new columns of values from the linked records.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> this = ibis.memtable({"record_id": [40, 50, 60], "x": ["a", "b", "c"]})
        >>> other = ibis.memtable({"record_id": [41, 42, 51], "y": [4.1, 4.2, 9.0]})
        >>> links = ibis.memtable(
        ...     {"record_id_l": [40, 40, 50], "record_id_r": [41, 42, 51]}
        ... )
        >>> lt = LinkedTable(this, other=other, links=links)
        >>> lt
        LinkedTable<3 records, 3 links>
        ┏━━━━━━━━━━━┳━━━━━━━━┓
        ┃ record_id ┃ x      ┃
        ┡━━━━━━━━━━━╇━━━━━━━━┩
        │ int64     │ string │
        ├───────────┼────────┤
        │        40 │ a      │
        │        50 │ b      │
        │        60 │ c      │
        └───────────┴────────┘

        Select exactly which values you want from `other`
        Since every record in `self` can be linked to 0-N records in `other`,
        we need a way of aggregating the 0-N linked values in `other` to a single value.
        By default, if the given expression is not an aggregate,
        we default to `.collect()` to combine them all into an array.
        If the given expression is already an aggregate, then it is kept as-is.

        This uses the same semantics as `ibis.Table.select(*values, **named_values)`.

        >>> lt.with_linked_values(
        ...     "y",
        ...     ibis._.y.max().name("y_max"),
        ...     others="record_id",
        ...     other=ibis._.record_id.arbitrary(),
        ... )
        LinkedTable<3 records, 3 links>
        ┏━━━━━━━━━━━┳━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━┓
        ┃ record_id ┃ x      ┃ y                    ┃ y_max   ┃ others               ┃ other ┃
        ┡━━━━━━━━━━━╇━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━┩
        │ int64     │ string │ array<float64>       │ float64 │ array<int64>         │ int64 │
        ├───────────┼────────┼──────────────────────┼─────────┼──────────────────────┼───────┤
        │        40 │ a      │ [4.1, 4.2]           │     4.2 │ [41, 42]             │    41 │
        │        50 │ b      │ [9.0]                │     9.0 │ [51]                 │    51 │
        │        60 │ c      │ NULL                 │    NULL │ NULL                 │  NULL │
        └───────────┴────────┴──────────────────────┴─────────┴──────────────────────┴───────┘

        One common use case for this is during the task of a lookup.
        I want to see the record id of the linked record,
        exluding when there are no links and when there are multiple (ie ambiguous) links.

        >>> lt.with_n_links().with_linked_values(
        ...     ibis._.record_id.arbitrary().name("record_id_other"),
        ... ).filter(ibis._.n_links == 1)
        ┏━━━━━━━━━━━┳━━━━━━━━┳━━━━━━━━━┳━━━━━━━━━━━━━━━━━┓
        ┃ record_id ┃ x      ┃ n_links ┃ record_id_other ┃
        ┡━━━━━━━━━━━╇━━━━━━━━╇━━━━━━━━━╇━━━━━━━━━━━━━━━━━┩
        │ int64     │ string │ int64   │ int64           │
        ├───────────┼────────┼─────────┼─────────────────┤
        │        50 │ b      │       1 │              51 │
        └───────────┴────────┴─────────┴─────────────────┘

        If no values are provided, we will by default add a column named `other`
        with all the values from other packed into an array of structs.

        >>> lt.with_linked_values()
        LinkedTable<3 records, 3 links>
        ┏━━━━━━━━━━━┳━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ record_id ┃ x      ┃ other                                                      ┃
        ┡━━━━━━━━━━━╇━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ int64     │ string │ array<struct<record_id: int64, y: float64>>                │
        ├───────────┼────────┼────────────────────────────────────────────────────────────┤
        │        40 │ a      │ [{'record_id': 41, 'y': 4.1}, {'record_id': 42, 'y': 4.2}] │
        │        50 │ b      │ [{'record_id': 51, 'y': 9.0}]                              │
        │        60 │ c      │ NULL                                                       │
        └───────────┴────────┴────────────────────────────────────────────────────────────┘
        """  # noqa: E501
        uname = _util.unique_name()
        if not values and not named_values:
            values = (
                lambda t: ibis.struct({c: t[c] for c in t.columns if c != uname}).name(
                    "other"
                ),
            )

        id_to_other = self.links_.select(
            **{
                uname: "record_id_l",
                "record_id": "record_id_r",
            }
        ).join(self.other_, "record_id")
        vals_bound = id_to_other.bind(*values, **named_values)
        aggs = [
            v if isinstance(v, ir.Scalar) else default_agg.resolve(v).name(v.get_name())
            for v in vals_bound
        ]

        self_id_to_other_array_vals = id_to_other.group_by(uname).agg(*aggs)
        value_names = [c for c in self_id_to_other_array_vals.columns if c != uname]

        t = self.mutate(_.record_id.name(uname))
        collisions = [c for c in value_names if c in t.columns]
        if collisions:
            raise ValueError(
                "Given values would collide with existing columns in self: "
                f"{collisions}",
            )
        with_other = _util.join_lookup(
            t,
            self_id_to_other_array_vals,
            uname,
        )
        with_other = with_other.select(*self.columns, *value_names)
        return self.__class__(with_other, other=self.other_, links=self.links_)

    def with_n_links(self, /, *, name: str = "n_links") -> _typing.Self:
        """
        Add a column to this table with the number of links each record has.

        Parameters
        ----------
        name
            The name of the new column.

        Returns
        -------
        A new LinkedTable with the new column.

        Examples
        --------
        >>> import ibis
        >>> import mismo
        >>> ibis.options.interactive = True
        >>> left = ibis.memtable({"record_id": [4, 5, 6]})
        >>> right = ibis.memtable({"record_id": [7, 8, 9]})
        >>> links = ibis.memtable({"record_id_l": [4, 4, 5], "record_id_r": [7, 8, 9]})
        >>> linkage = mismo.Linkage(left=left, right=right, links=links)
        >>> linkage.left.with_n_links().order_by("record_id")
        ┏━━━━━━━━━━━┳━━━━━━━━━┓
        ┃ record_id ┃ n_links ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━┩
        │ int64     │ int64   │
        ├───────────┼─────────┤
        │         4 │       2 │
        │         5 │       1 │
        │         6 │       0 │
        └───────────┴─────────┘
        >>> linkage.right.with_n_links(name="link_count").order_by("record_id")
        ┏━━━━━━━━━━━┳━━━━━━━━━━━━┓
        ┃ record_id ┃ link_count ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━━━━┩
        │ int64     │ int64      │
        ├───────────┼────────────┤
        │         7 │          1 │
        │         8 │          1 │
        │         9 │          1 │
        └───────────┴────────────┘
        """
        # this doesn't include counts for ids that aren't in links
        # (eg those that aren't linked at all)
        # So make sure these get added back in during the join_lookup()
        n_by_id = self.links_.group_by(record_id="record_id_l").aggregate(
            _.record_id_r.nunique().name(name)
        )
        t = self
        if name in t.columns:
            t = t.drop(name)
        added = _util.join_lookup(t, n_by_id, "record_id", defaults={name: 0})
        return self.__class__(added, other=self.other_, links=self.links_)

    def link_counts(self) -> LinkCountsTable:
        """
        Describes 'There are `n_records` in self that linked to `n_links` in `other'.

        This is basically a histogram of `self.with_n_links()`

        See Also
        --------
        with_n_links

        Examples
        --------
        >>> import ibis
        >>> import mismo
        >>> ibis.options.interactive = True
        >>> left = ibis.memtable({"record_id": [4, 5, 6]})
        >>> right = ibis.memtable({"record_id": [7, 8, 9]})
        >>> links = ibis.memtable({"record_id_l": [4, 4, 5], "record_id_r": [7, 8, 9]})
        >>> linkage = mismo.Linkage(left=left, right=right, links=links)

        There is 1 record in left (6) that didn't match any in right.
        There is 1 record in left (5) that matched 1 in right.
        There is 1 record in left (4) that matched 2 in right.

        >>> linkage.left.link_counts().order_by("n_links")
        ┏━━━━━━━━━┳━━━━━━━━━━━┓
        ┃ n_links ┃ n_records ┃
        ┡━━━━━━━━━╇━━━━━━━━━━━┩
        │ int64   │ int64     │
        ├─────────┼───────────┤
        │       0 │         1 │
        │       1 │         1 │
        │       2 │         1 │
        └─────────┴───────────┘

        All 3 records in right matched 1 in left.

        >>> linkage.right.link_counts()
        ┏━━━━━━━━━┳━━━━━━━━━━━┓
        ┃ n_links ┃ n_records ┃
        ┡━━━━━━━━━╇━━━━━━━━━━━┩
        │ int64   │ int64     │
        ├─────────┼───────────┤
        │       1 │         3 │
        └─────────┴───────────┘
        """
        basic_counts = self.links_.group_by(record_id="record_id_l").aggregate(
            n_links=_.record_id_r.nunique()
        )
        zero_counts = self.filter(_.record_id.notin(basic_counts.record_id)).select(
            "record_id", n_links=ibis.literal(0, type="int64")
        )
        n_links_by_id = ibis.union(basic_counts, zero_counts)
        counts = n_links_by_id.group_by("n_links").aggregate(
            n_records=_.record_id.count()
        )
        return LinkCountsTable(counts)

    @classmethod
    def make_pair(
        cls, *, left: ir.Table, right: ir.Table, links: ir.Table
    ) -> tuple[_typing.Self, _typing.Self]:
        """
        Create a pair of LinkedTables from left, right, and links.

        This basically just wraps the logic to make it so that
        the the _l and _r suffixes in the links table are consistent.
        """

        left = cls(left, other=right, links=links)
        right = cls(right, other=left, links=LinksTable._swap_perspective(links))
        return left, right

    def __repr__(self) -> str:
        return f"""{self.__class__.__name__}<{self.count().execute()} records, {self.links_.count().execute()} links>
{super().__repr__()}
""".strip()  # noqa: E501


class LinkCountsTable(TableWrapper):
    """A table representing the number of records binned by number of links.

    eg "There were 700 records with 0 links, 300 with 1 link, 20 with 2 links, ..."
    """

    n_records: ir.IntegerColumn
    """The number of records."""
    n_links: ir.IntegerColumn
    """The number of links."""

    def __init__(self, t: ibis.Table) -> None:
        """Create from an ibis table with exactly columns 'n_records' and 'n_links'.

        Parameters
        ----------
        t
            The table with exactly columns 'n_records' and 'n_links'.
        """
        if set(t.columns) != {"n_records", "n_links"}:
            raise ValueError(
                "LinkCountsTable must have exactly columns 'n_records' and 'n_links'"
            )
        super().__init__(t)

    def chart(self) -> alt.Chart:
        """A bar chart of the number of records by the number of links.

        ```plaintext
                         Number of Records
        Number of          By Link Count
          Records
                |    █
        100,000 | █  █
                | █  █
                | █  █  █
         10,000 | █  █  █
                | █  █  █
                | █  █  █
          1,000 | █  █  █  █
                | █  █  █  █  █  █
                | █  █  █  █  █  █  █
            100 | █  █  █  █  █  █  █  █  █
                | 0  1  2  3  4 10 12 14 23
                Number of Links
        ```
        """
        import altair as alt

        total_records = self.n_records.sum().fill_null(0).execute()
        n_title = ["Number of Records", f"({total_records:_} total)"]
        key_title = "Number of Links"
        mins = self.order_by(_.n_links.asc()).limit(2).execute()
        if len(mins) > 1:
            subtitle = f"eg '{mins.n_records[0]:_} records had {mins.n_links[0]:_} links, {mins.n_records[1]:_} had {mins.n_links[1]:_} links, ...'"  # noqa: E501
        elif len(mins) == 1:
            subtitle = (
                f"eg '{mins.n_records[0]:_} records had {mins.n_links[0]:_} links', ..."  # noqa: E501
            )
        else:
            subtitle = "eg 'there were 1000 records with 0 links, 500 with 1 link, 100 with 2 links, ...'"  # noqa: E501

        frac_records: ir.StringValue = cast(
            ir.StringValue,
            (self.n_records / total_records * 100).cast(int).cast(str)
            if total_records > 0
            else ibis.literal("0"),
        )
        t = self.mutate(
            frac_records=frac_records,
            explanation=(
                "Out of the "
                + f"{total_records:_} records, "
                + self.n_records.cast(str)
                + " ("
                + frac_records
                + "%) had "
                + self.n_links.cast(str)
                + " links."
            ),
        )
        scrubber_selection = alt.selection_interval(
            encodings=["x"],
            empty=True,
            # Ensure a unique name to avoid conflicts with other LinkCountsTable charts.
            # https://github.com/vega/altair/issues/3891
            name=_util.unique_name(),
        )
        width = 800
        zoomin = (
            alt.Chart(t, width=width)
            .mark_bar()
            .encode(
                # if we ever change this sorting, keep the subtitle example
                # to be in sync of the same order as the bars go left to right.
                alt.X("n_links:O", title=key_title, sort="x"),
                alt.Y(
                    "n_records:Q",
                    title=n_title,
                    scale=alt.Scale(type="symlog"),
                ),
                tooltip=[
                    alt.Tooltip("n_records:Q", title=n_title, format=","),
                    alt.Tooltip(
                        "frac_records:Q", title="Fraction of Records", format=".2%"
                    ),
                    alt.Tooltip("n_links:O", title=key_title),
                    alt.Tooltip("explanation", title="Explanation"),
                ],
            )
            .transform_filter(scrubber_selection)
        )
        scrubber = (
            alt.Chart(
                t,
                title=alt.Title(
                    text="<Drag to select>",
                    dy=30,
                    anchor="middle",
                    fontSize=12,
                    color="gray",
                ),
                width=width,
                height=50,
            )
            .mark_area(interpolate="step-after")
            .encode(
                alt.X("n_links:O", sort="x", axis=None),
                alt.Y("n_records:Q", title=None, axis=None),
            )
            .add_params(scrubber_selection)
        )
        together = scrubber & zoomin
        together = together.resolve_scale(color="independent")
        together = together.properties(
            title=alt.Title(
                "Number of Records by Link Count",
                subtitle=subtitle,
                anchor="middle",
                fontSize=14,
            ),
        )
        return together



from __future__ import annotations

from json import dumps, loads
import math
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Iterator, overload

import ibis
from ibis.expr import types as ir

from mismo import _util
from mismo.compare import LevelComparer

from .._typing import Self
from ._util import odds_to_log_odds, odds_to_prob

if TYPE_CHECKING:
    import altair as alt


class LevelWeights:
    """Weights for a single [MatchLevel][mismo.compare.MatchLevel].

    This describes for example "If zipcodes match perfectly, then
    this increases the probability of a match by 10x as compared to if we
    hadn't looked at zipcode".
    """

    def __init__(self, name: str, *, m: float, u: float) -> None:
        """Create a new LevelWeights object."""
        self._name = name
        self._m = m
        self._u = u

    @property
    def name(self) -> str:
        """The name of the level, e.g. "Exact Match"."""
        return self._name

    @property
    def m(self) -> float:
        """Among true-matches, what proportion of them have this level?

        1 means this level is a good indication of a match, 0 means it's a good
        indication of a non-match.
        """
        return self._m

    @property
    def u(self) -> float:
        """Among non-matches, what proportion of them have this level?

        1 means this level is a good indication of a non-match, 0 means it's a good
        indication of a match.
        """
        return self._u

    @property
    def odds(self) -> float:
        """How much more likely is a match than a non-match at this level?

        This is derived from m and u. This is the same thing as "Bayes Factor"
        in splink.

        - values below 1 is evidence against a match
        - values above 1 is evidence for a match
        - 1 means this level does not provide any evidence for or against a match
        """
        if self.u == 0:
            return float("inf")
        else:
            return self.m / self.u

    @property
    def log_odds(self) -> float:
        """The log base 10 of the odds."""
        return odds_to_log_odds(self.odds)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, m={self.m}, u={self.u})"

    def __eq__(self, other):
        if not isinstance(other, LevelWeights):
            return False
        return (
            self.name == other.name
            and math.isclose(self.m, other.m)
            and math.isclose(self.u, other.u)
        )


class ComparerWeights:
    """
    The weights for a single [LevelComparer][mismo.compare.LevelComparer].

    An ordered, dict-like collection of [LevelWeights][mismo.fs.LevelWeights]
    one for each level.
    """

    def __init__(self, name: str, level_weights: Iterable[LevelWeights]):
        """Create a new ComparerWeights object."""
        self._name = name
        self._level_weights = tuple(level_weights)
        lookup = {}
        for i, lw in enumerate(self._level_weights):
            if lw.name in lookup:
                raise ValueError(f"Duplicate level name: {lw.name}")
            # 0 -> -len + 0
            # 1 -> -len + 1
            # 2 -> -len + 2
            negative_i = -len(self._level_weights) + i
            lookup[lw.name] = lw
            lookup[i] = lw
            lookup[negative_i] = lw
        self._lookup = lookup

    @property
    def name(self) -> str:
        """
        The name of the LevelComparer these weights are for, eg 'name" or "address".
        """
        return self._name

    @overload
    def __getitem__(self, name_or_index: str | int) -> LevelWeights: ...

    @overload
    def __getitem__(self, name_or_index: slice) -> tuple[LevelWeights, ...]: ...

    def __getitem__(
        self, name_or_index: str | int | slice
    ) -> LevelWeights | tuple[LevelWeights, ...]:
        """Get a LevelWeights by name or index."""
        if isinstance(name_or_index, slice):
            return self._level_weights[name_or_index]
        try:
            return self._lookup[name_or_index]
        except KeyError:
            raise KeyError(f"Unknown level name or index: {name_or_index}")

    def __contains__(self, name_or_index: str | int) -> bool:
        """Check if a LevelWeights is present by name or index."""
        return name_or_index in self._lookup

    def __iter__(self) -> Iterator[LevelWeights]:
        """Iterate over the LevelWeights, including the implicit ELSE level."""
        return iter(self._level_weights)

    def __len__(self) -> int:
        """The number of LevelWeights, including the implicit ELSE level."""
        return len(self._level_weights)

    def __repr__(self) -> str:
        return f"""{self.__class__.__name__}(
        name={self.name},
        level_weights={self._level_weights}
    )"""

    def __eq__(self, other):
        if not isinstance(other, ComparerWeights):
            return False
        return self.name == other.name and self._level_weights == other._level_weights

    @overload
    def odds(self, labels: str | int) -> float: ...

    @overload
    def odds(self, labels: ir.StringValue | ir.IntegerValue) -> ir.FloatingValue: ...

    def odds(
        self, labels: str | int | ir.StringValue | ir.IntegerValue
    ) -> float | ir.FloatingValue:
        """Calculate the odds for each record pair.

        If `labels` is a string or integer, then we calculate the odds for that
        level. For example, if `labels` is "close", then we calculate the odds
        for the "close" level. If `labels` is 0, then we calculate the odds for
        the first level. If `labels` is -1, then we calculate the odds for the
        last level (the ELSE level).

        If `labels` is a StringValue or IntegerValue, then we do the same thing,
        except that we return an ibis FloatingValue instead of a python float.
        """
        if isinstance(labels, (int, str)):
            return self[labels].odds
        if isinstance(labels, ir.StringValue):
            cases = [(labels == lw.name, lw.odds) for lw in self]
        elif isinstance(labels, ir.IntegerValue):
            cases = [(labels == i, lw.odds) for i, lw in enumerate(self)]  # type: ignore # noqa: E501
        else:
            raise TypeError(
                f"Expected int, str, StringValue, or IntegerValue, got {type(labels)}"
            )
        return _util.cases(*cases)

    @overload
    def match_probability(self, labels: str | int) -> float: ...

    @overload
    def match_probability(
        self, labels: ir.StringValue | ir.IntegerValue
    ) -> ir.FloatingValue: ...

    def match_probability(
        self, labels: str | int | ir.StringValue | ir.IntegerValue
    ) -> float | ir.FloatingValue:
        """Calculate the match probability for each record pair."""
        return odds_to_prob(self.odds(labels))

    @overload
    def log_odds(self, labels: str | int) -> float: ...

    @overload
    def log_odds(
        self, labels: ir.StringValue | ir.IntegerValue
    ) -> ir.FloatingValue: ...

    def log_odds(
        self, labels: str | int | ir.StringValue | ir.IntegerValue
    ) -> float | ir.FloatingValue:
        """Calculate the log odds for each record pair."""
        return odds_to_log_odds(self.odds(labels))

    @staticmethod
    def plot(self) -> alt.Chart:
        """Plot the weights for this Comparer."""
        from ._plot import plot_weights

        return plot_weights(self)


def compare_one(
    t: ir.Table, level_comparer: LevelComparer, comp_weights: ComparerWeights
) -> tuple[ir.StringValue, ir.FloatingValue]:
    conditions = [level.is_match(t) for level in level_comparer]
    odds = [level_weights.odds for level_weights in comp_weights]
    return (
        _util.cases(*zip(conditions, level_comparer.levels), else_="else").name(
            comp_weights.name
        ),
        _util.cases(*zip(conditions, odds), else_=1).name(f"{comp_weights.name}_odds"),
    )


class Weights:
    """Weights for the Fellegi-Sunter model.

    An unordered, dict-like collection of
    [ComparerWeights][mismo.fs.ComparerWeights],
    one for each [LevelComparer][mismo.compare.LevelComparer] of the same name.
    """

    def __init__(self, comparer_weights: Iterable[ComparerWeights]):
        """Create a new Weights object."""
        self._lookup = {cw.name: cw for cw in comparer_weights}

    def __getitem__(self, name: str) -> ComparerWeights:
        """Get a `ComparerWeights` by name."""
        return self._lookup[name]

    def __iter__(self) -> Iterator[ComparerWeights]:
        """Iterate over the contained `ComparerWeights`."""
        return iter(self._lookup.values())

    def __len__(self) -> int:
        """The number of `ComparerWeights`."""
        return len(self._lookup)

    def score_compared(self, compared: ir.Table) -> ir.Table:
        """Score already-compared record pairs.

        This assumes that there is already a column one for each LevelComparer
        that contains the labels for each record pair. For example, if we have
        a LevelComparer called "address", then we should have a column called
        "address" that contains labels like "exact", "one-letter-off", "same-city", etc.

        For each LevelComparer, we add a column, `{comparer.name}_odds`.
        This is a number that describes how this comparer affects the likelihood
        of a match. For example, an odds of 10 means that this comparer
        increased the likelihood of a match by 10x as compared to if we hadn't
        looked at this comparer.
        For example, the column might be called "name_odds" and have values like
        10, 0.1, 1.

        In addition to these per-LevelComparer columns, we also add a column
        called "odds" which is the overall odds for each record pair.
        We calculate this by starting with the odds of 1 and then multiplying
        by each LevelComparer's odds to get the overall odds.
        """
        results = []
        for comparer_weights in self:
            name = comparer_weights.name
            labels = compared[name]
            odds = comparer_weights.odds(labels)
            results.append((name, labels, odds))
        return self._score(compared, results)

    def compare_and_score(
        self, t: ir.Table, level_comparers: Iterable[LevelComparer]
    ) -> ir.Table:
        """Compare and score record pairs.

        Use the given `level_comparers` to label the record pairs, and then
        score the results using `self.score_compared`.
        """
        results = []
        for cmp, weights in zip(level_comparers, self):
            label, odds = compare_one(t, cmp, weights)
            results.append((cmp.name, label, odds))
        return self._score(t, results)

    def _score(self, t: ir.Table, compare_results) -> ir.Table:
        # workaround for https://github.com/ibis-project/ibis/issues/10890
        # to ensure that duckdb treats the literal as a float64
        total_odds = ibis.literal("1").cast("float64")
        m = {}
        naming = {}
        for name, label, odds in compare_results:
            m[name] = label
            m[f"{name}_odds"] = odds
            total_odds *= odds
            naming[f"{name}_odds"] = name
        m["odds"] = total_odds
        result = t.mutate(**m)
        # Don't do any of this relocation in terms of record_id, etc
        # because the passed in table doesn't need ot have these.
        # It only needs to have the labels for each comparer.
        for odds_name, name in naming.items():
            result = result.relocate(odds_name, after=name)
        result = result.relocate("odds", before=list(naming.values())[0])
        return result

    def plot(self) -> alt.Chart:
        """Plot the weights for all of the LevelComparers."""
        from ._plot import plot_weights

        return plot_weights(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{tuple(self)}"

    def __eq__(self, other):
        if not isinstance(other, Weights):
            return False
        a = sorted(self, key=lambda cw: cw.name)
        b = sorted(other, key=lambda cw: cw.name)
        return a == b

    def to_json(self, path: str | Path | None = None) -> dict:
        """Return a JSON-serializable representation of the weights.

        If `path` is given, write the dict to the file at that path in addition
        to returning it.
        """
        d = {
            cw.name: {lw.name: {"m": lw.m, "u": lw.u} for lw in cw if lw.name != "else"}
            for cw in self
        }
        if path is not None:
            Path(path).write_text(dumps(d))
        return d

    @classmethod
    def from_json(cls, json: dict | str | Path) -> Self:
        """Create a Weights object from a JSON-serializable representation.

        Parameters
        ----------
        json:
            If a dict, assumed to be the JSON-serializable representation.
            Load it directly.
            If a str or Path, assumed to be a path to a JSON file.
            Load it from that file.

        Returns
        -------
        Weights
            The Weights object created from the JSON-serializable representation.
        """
        if not isinstance(json, dict):
            json = loads(Path(json).read_text())
        return cls(
            ComparerWeights(
                name=comparer_name,
                level_weights=[
                    LevelWeights(name=level_name, m=level["m"], u=level["u"])
                    for level_name, level in comparer.items()
                ],
            )
            for comparer_name, comparer in json.items()
        )
